'''
 2019 Bjoern Annighoefer
'''

class CmdTypes:
    #data related commmands
    GET = 'GET' #get cmd
    SET = 'SET' #set cmd value
    ADD = 'ADD' #add cmd value
    REM = 'REM' #remove cmd value
    MOV = 'MOV' #move cmd cmd
    DEL = 'DEL' #delete cmd
    CLO = 'CLO' #clone source target mode
    CRT = 'CRT' #create by class
    CRN = 'CRN' #create by name
    QRF = 'QRF' #querify
    
    #comparison and merge
    CPR = 'CPR' #compare/diff
    MRG = 'MRG' #merge 
    
    #meta model related commands
    GMM = 'GMM' #get meta models
    RMM = 'RMM' #register meta model
    UMM = 'UMM' #unregister meta model
    
    #maintenance related commands
    HEL = 'HEL' #hello
    GBY = 'GBY' #goodbye
    SES = 'SES' #session
    STS = 'STS' #status
    CHG = 'CHG' #changes
    OBS = 'OBS' #observe
    UBS = 'UBS' #unobserve
    UDO = 'UDO' #undo
    RDO = 'RDO' #redo
    
    # Action related commands
    GAA = 'GAA' #get all actions
    CAL = 'CAL' #call
    ASC = 'ASC' #async Call
    ABC = 'ABC' #abort call
    CST = 'CST' #call status

    # Validation related commands
    ADC = "ADC"
    RMC = "RMC"
    VAL = "VAL"
    GAC = "GAC"
    
    CMP = 'CMP' #compound
    MUT = 'MUT' #mute stop outputing results history only
    UMT = 'UMT' #unmute start outputing results
    
    
class CloModes:
    CLS = 'CLS' #class: object only
    ATT = 'ATT' #attribute: class + attributes
    FLT = 'FLT' #deep: class + attributes + references
    DEP = 'DEP' #deep: classes + attributes + containments
    FUL = 'FUL' #full: classes + attributes + containments + reference adaptation

class MrgModes:
    STD = 'STD' #standard mode without additional information or output
    LST = 'LST' #list: creates a list with all commands
    LOG = 'LOG' #log: LST + creates a log file (info.log) with detailed information on every object
    DET = 'DET' #detailed: LOG + creates a file with all objects and their #1234 number

class Cmd:
    def __init__(self,t : str,args):
        self.cmd = t
        self.a = args
    def __repr__(self):
        cmdStr = str(self.cmd)
        if(type(self.a)==list and len(self.a) > 0):
            cmdStr += ' ' + ' '.join([str(arg) for arg in self.a])
        elif(self.a):
            cmdStr += ' ' + str(self.a)
        return cmdStr
        
class Get(Cmd):
    def __init__(self,target):
        super().__init__(CmdTypes.GET, target)
        
class Set(Cmd):
    def __init__(self,target,feature,value):
        super().__init__(CmdTypes.SET, [target,feature,value])
        
class Add(Cmd):
    def __init__(self,target,feature,value):
        super().__init__(CmdTypes.ADD, [target,feature,value])
        
class Rem(Cmd):
    def __init__(self,target,feature,value):
        super().__init__(CmdTypes.REM, [target,feature,value])
        
class Mov(Cmd):
    def __init__(self,target,newIndex):
        super().__init__(CmdTypes.MOV, [target,newIndex])
        
class Clo(Cmd):
    def __init__(self,target,original):
        super().__init__(CmdTypes.CLO,[target,original])
        
class Crt(Cmd):
    def __init__(self,clazz,n,constructorArgs=[]):
        super().__init__(CmdTypes.CRT,[clazz,n,constructorArgs])
        
class Crn(Cmd):
    def __init__(self,package,name,n,constructorArgs=[]):
        super().__init__(CmdTypes.CRN,[package,name,n,constructorArgs])
        
class Qrf(Cmd):
    def __init__(self,target):
        super().__init__(CmdTypes.QRF,target)
        
class Cpr(Cmd):
    def __init__(self,old,new,mode):
        super().__init__(CmdTypes.CPR,[old,new,mode])
        
class Mrg(Cmd):
    def __init__(self,old,new,mode):
        super().__init__(CmdTypes.MRG,[old,new,mode])

class Gmm(Cmd):
    def __init__(self):
        super().__init__(CmdTypes.GMM, None)
        
class Rmm(Cmd):
    def __init__(self,metamodel):
        super().__init__(CmdTypes.RMM, metamodel)
        
class Umm(Cmd):
    def __init__(self,metamodel):
        super().__init__(CmdTypes.UMM, metamodel)
        
class Hel(Cmd):
    def __init__(self,user,password):
        super().__init__(CmdTypes.HEL, [user, password])
        
class Ses(Cmd):
    def __init__(self,sessionId):
        super().__init__(CmdTypes.SES, sessionId)
        
class Gby(Cmd):
    def __init__(self,sessionId):
        super().__init__(CmdTypes.GBY, sessionId)
        
class Sts(Cmd):
    def __init__(self):
        super().__init__(CmdTypes.STS, None)
        
class Chg(Cmd):
    def __init__(self,latestChangeId,n):
        super().__init__(CmdTypes.CHG,[latestChangeId,n])
        
class Obs(Cmd):
    def __init__(self,eventType,eventKey):
        super().__init__(CmdTypes.OBS, [eventType,eventKey])
        
class Ubs(Cmd):
    def __init__(self,eventType,eventKey):
        super().__init__(CmdTypes.UBS, [eventType,eventKey])
        
class Gaa(Cmd):
    def __init__(self):
        super().__init__(CmdTypes.GAA,None)
        
class Cal(Cmd):
    def __init__(self,name,args=[],opts=[]):
        super().__init__(CmdTypes.CAL,[name,args,opts])
    
class Asc(Cmd):
    def __init__(self,name,args=[],opts=[]):
        super().__init__(CmdTypes.ASC,[name,args,opts])
        
class Abc(Cmd):
    def __init__(self,callId):
        super().__init__(CmdTypes.ABC,callId)

class Adc(Cmd):
    def __init__(self, constraintType, target, feature, law, name, annotation):
        super().__init__(CmdTypes.ADC,[constraintType, target, feature, law, name, annotation])

class Rmc(Cmd):
    def __init__(self, constraint):
        super().__init__(CmdTypes.RMC,constraint)

class Val(Cmd):
    def __init__(self, obj):
        super().__init__(CmdTypes.VAL, obj)

class Gac(Cmd):
    def __init__(self):
        super().__init__(CmdTypes.GAC, None)

class Mut(Cmd):
    def __init__(self):
        super().__init__(CmdTypes.MUT,None)
        
class Umt(Cmd):
    def __init__(self):
        super().__init__(CmdTypes.UMT,None)
        
class Cmp(Cmd):
    def __init__(self):
        super().__init__(CmdTypes.CMP,[])
        
    def Get(self,target):
        self.a.append(Get(target))
        return self
    
    def Set(self,target,feature,value):
        self.a.append(Set(target,feature,value))
        return self
    
    def Add(self,target,feature,value):
        self.a.append(Add(target,feature,value))
        return self
    
    def Rem(self,target,feature,value):
        self.a.append(Rem(target,feature,value))
        return self
    
    def Mov(self,target,newIndex):
        self.a.append(Mov(target,newIndex))
        return self
    
    def Clo(self,target,mode):
        self.a.append(Clo(target,mode))
        return self
    
    def Crt(self,clazz,n,constructorArgs=[]):
        self.a.append(Crt(clazz,n,constructorArgs))
        return self
    
    def Crn(self,package,name,n,constructorArgs=[]):
        self.a.append(Crn(package,name,n,constructorArgs))
        return self
    
    def Qrf(self,target):
        self.a.append(Qrf(target))
        return self
    
    def Mrg(self,old,new,mode):
        self.a.append(Mrg(old,new,mode))
        return self

    def Cpr(self,old,new,mode):
        self.a.append(Cpr(old,new,mode))
        return self
    
    def Gmm(self):
        self.a.append(Gmm())
        return self
    
    def Rmm(self,metamodel):
        self.a.append(Rmm(metamodel))
        return self
    
    def Umm(self,metamodel):
        self.a.append(Umm(metamodel))
        return self
    
    def Hel(self,user,password):
        self.a.append(Hel(user,password))
        return self
    
    def Ses(self,sessionId):
        self.a.append(Ses(sessionId))
        return self
    
    def Gby(self,sessionId):
        self.a.append(Gby(sessionId))
        return self
    
    def Sts(self):
        self.a.append(Sts())
        return self
    
    def Chg(self,changeId,n):
        self.a.append(Chg(changeId,n))
        return self
    
    def Obs(self,eventType,eventKey):
        self.a.append(Obs(eventType,eventKey))
        return self
    
    def Ubs(self,eventType,eventKey):
        self.a.append(Ubs(eventType,eventKey))
        return self
    
    def Gaa(self):
        self.a.append(Gaa())
        return self
    
    def Cal(self,name,args=[],opts=[]):
        self.a.append(Cal(name,args,opts))
        return self
    
    def Asc(self,name,args=[],opts=[]):
        self.a.append(Asc(name,args,opts))
        return self
    
    def Abc(self,callId):
        self.a.append(Abc(callId))
        return self

    def Adc(self,constraintType, target, feature, law, name, annotation):
        self.a.append(Adc(constraintType, target, feature, law, name, annotation))
        return self

    def Rmc(self,constraint):
        self.a.append(Rmc(constraint))
        return self

    def Val(self,obj):
        self.a.append(Val(obj))
        return self

    def Gac(self):
        self.a.append(Gac())
        return self
    
    def Mut(self):
        self.a.append(Mut())
        return self
    
    def Umt(self):
        self.a.append(Umt())
        return self



from ..model import DomainA, CommandA, ResultA, ErrorResult
from .. import CommandParser
from .. import ResultParser

from multiprocessing import Queue
import queue # imported for using queue.Empty exception

class MultiprocessingQueueDomainClient(DomainA):
    def __init__(self,cmdQueue : Queue,resQueue : Queue,timeout=1000000):
        self.cmdQueue = cmdQueue
        self.resQueue = resQueue
        self.timeout = timeout
        
    def Do(self, command : CommandA) -> ResultA:
        cmdStr = CommandParser.CommandToString(command)
        self.cmdQueue.put(cmdStr)
        try:
            resultStr = self.resQueue.get(timeout=self.timeout)
            result = ResultParser.StringToResult(resultStr)
        except queue.Empty:
            result = ErrorResult()
            result.code = 9880
            result.message = "Internal multiprocessing queue receive timout after %s s."%(self.timout)
        
        return result
        




'''
 Bjoern Annighoefer 2019
'''

from .serializer import Serializer

from ..query.query import QrySegTypes
from ..command.command import CmdTypes
from ..util.error import EoqError



'''
TEXT Serializer   

'''

class PySerializer(Serializer):
    def __init__(self):
        self.cmdTranslators = {
            CmdTypes.GET : lambda o: "Get("+self.Ser(o.a)+")",
            CmdTypes.SET : lambda o: "Set("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.ADD : lambda o: "Add("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.REM : lambda o: "Rem("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.MOV : lambda o: "Mov("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.CLO : lambda o: "Clo("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.CRN : lambda o: "Crn("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.CRT : lambda o: "Crt("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.QRF : lambda o: "Qrf("+self.Ser(o.a)+")",
            CmdTypes.CPR : lambda o: "Cpr("+self.Ser(o.a[0])+","+self.Ser(o.a[1])+","+self.Ser(o.a[2])+")",
            CmdTypes.MRG : lambda o: "Mrg("+self.Ser(o.a[0])+","+self.Ser(o.a[1])+","+self.Ser(o.a[2])+")",
            CmdTypes.GMM : lambda o: "Gmm()",
            CmdTypes.RMM : lambda o: "Rmm("+self.Ser(o.a)+")",
            CmdTypes.UMM : lambda o: "Umm("+self.Ser(o.a)+")",
            CmdTypes.HEL : lambda o: "Hel("+self.Ser(o.a[0])+","+self.Ser(o.a[1])+")",
            CmdTypes.SES : lambda o: "Ses("+self.Ser(o.a)+")",
            CmdTypes.GBY : lambda o: "Gby("+self.Ser(o.a)+")",
            CmdTypes.STS : lambda o: "Sts()",
            CmdTypes.CHG : lambda o: "Chg("+",".join([self.Ser(a) for a in o.a])+")",
            CmdTypes.OBS : lambda o: "Obs("+self.Ser(o.a[0])+","+self.Ser(o.a[1])+")",
            CmdTypes.UBS : lambda o: "Ubs("+self.Ser(o.a[0])+","+self.Ser(o.a[1])+")",
            CmdTypes.GAA : lambda o: "Gaa()",
            CmdTypes.CAL : lambda o: "Cal("+self.Ser(o.a[0])+","+self.Ser(o.a[1:])+")",
            CmdTypes.ASC : lambda o: "Asc("+self.Ser(o.a[0])+","+self.Ser(o.a[1:])+")",
            CmdTypes.ABC : lambda o: "Abc("+self.Ser(o.a)+")",
            CmdTypes.CST : lambda o: "Cst("+self.Ser(o.a)+")",
            CmdTypes.CMP : lambda o: "Cmp()."+".".join([self.Ser(a) for a in o.a]),
            CmdTypes.MUT : lambda o: "Mut()",
            CmdTypes.UMT : lambda o: "Umt()"
            }
        self.qryTranslators = {
            QrySegTypes.OBJ : lambda o: "Obj("+self.Ser(o.v)+")",
            QrySegTypes.HIS : lambda o: "His("+self.Ser(o.v)+")",
            QrySegTypes.PTH : lambda o: "Pth("+self.Ser(o.v)+")",
            QrySegTypes.CLS : lambda o: "Cls("+self.Ser(o.v)+")",
            QrySegTypes.INO : lambda o: "Ino("+self.Ser(o.v)+")",
            QrySegTypes.MET : lambda o: "Met("+self.Ser(o.v[0]) + (","+self.Ser(o.v[1:]) if len(o.v)>1 else "")+")",
            QrySegTypes.NOT : lambda o: "Not()",
            QrySegTypes.TRM : lambda o: "Trm("+self.Ser(o.v[0])+","+self.Ser(o.v[1])+")",
            QrySegTypes.TRY : lambda o: "Try("+self.Ser(o.v[0])+","+self.Ser(o.v[1])+")",
            QrySegTypes.IDX : lambda o: "Idx("+self.Ser(o.v)+")",
            QrySegTypes.SEL : lambda o: "Sel("+self.Ser(o.v)+")",
            QrySegTypes.ARR : lambda o: "Arr(["+",".join([self.Ser(a) for a in o.v])+"])",
            QrySegTypes.ZIP : lambda o: "Zip("+self.Ser(o.v)+")",
            QrySegTypes.QRY : lambda o: "Qry()."+".".join([self.Ser(a) for a in o.v]),
            QrySegTypes.ANY : lambda o: "Any("+self.Ser(o.v)+")",
            QrySegTypes.ALL : lambda o: "All("+self.Ser(o.v)+")",
            QrySegTypes.EQU : lambda o: "Equ("+self.Ser(o.v)+")",
            QrySegTypes.EQA : lambda o: "Eqa("+self.Ser(o.v)+")",
            QrySegTypes.NEQ : lambda o: "Neq("+self.Ser(o.v)+")",
            QrySegTypes.LES : lambda o: "Les("+self.Ser(o.v)+")",
            QrySegTypes.GRE : lambda o: "Gre("+self.Ser(o.v)+")",
            QrySegTypes.RGX : lambda o: "Rgx("+self.Ser(o.v)+")",
            QrySegTypes.ADD : lambda o: "Add("+self.Ser(o.v)+")",
            QrySegTypes.SUB : lambda o: "Sub("+self.Ser(o.v)+")",
            QrySegTypes.MUL : lambda o: "Mul("+self.Ser(o.v)+")",
            QrySegTypes.DIV : lambda o: "Div("+self.Ser(o.v)+")",
            QrySegTypes.ORR : lambda o: "Orr("+self.Ser(o.v)+")",
            QrySegTypes.XOR : lambda o: "Xor("+self.Ser(o.v)+")",
            QrySegTypes.AND : lambda o: "And("+self.Ser(o.v)+")",
            QrySegTypes.NAD : lambda o: "Nad("+self.Ser(o.v)+")",
            QrySegTypes.CSP : lambda o: "Csp("+self.Ser(o.v)+")",
            QrySegTypes.ITS : lambda o: "Its("+self.Ser(o.v)+")",
            QrySegTypes.DIF : lambda o: "Dif("+self.Ser(o.v)+")",
            QrySegTypes.UNI : lambda o: "Uni("+self.Ser(o.v)+")",
            QrySegTypes.CON : lambda o: "Con("+self.Ser(o.v)+")"
            }
        self.priTranslators = {
            #primitive types
            bool : lambda o: str(o),
            int : lambda o: str(o),
            float : lambda o: str(o),
            bool : lambda o: str(o),
            str : lambda o: "'"+o+"'",
            list : lambda o: "[%s]"%(",".join([self.Ser(x) for x in o])),
            type(None) : lambda o: str(o)
            }

    def Ser(self,val):
        try:
            return self.cmdTranslators[val.cmd](val)
        except:
            try:
                return self.qryTranslators[val.qry](val)
            except:
                try:
                    return self.priTranslators[type(val)](val)
                except:
                    raise EoqError(0,"Text serializer failed for %s"%(str(val)))

    def Des(self,code):
        pass


'''
 Bjoern Annighoefer 2019
'''

#Abstract definition of a domain
class Mdb:     
    def Root(self):
        pass
    
    def Metamodels(self):
        pass
    
    def AddMetamodel(self,name,metamodel):
        pass
        
    def RemoveMetamodel(self,name,metamodel):
        pass
        
    def GetMetamodel(self,name):
        pass
    
    def ResolveProxy(self,proxy):
        pass
    
    def Close(self):
        pass
    
    


'''
 Bjoern Annighoefer 2019
 Modified: Christian Mollière 2020
'''

from .serializer import Serializer

from ..query.query import QRY_SYMBOLS, QrySegTypes, Qry, Obj, His, Cls, Ino, Met, Idx, Pth, Arr, Equ, Eqa, Neq, Les, \
    Gre, Rgx, Not, Try, Zip, All, Any
from ..command.command import CmdTypes, Get, Set, Add, Rem, Mov, Clo, Crt, Crn, Qrf, Sts, Chg, Gaa, Cal, Asc, Abc, Cmp, Umm, \
    Rmm, Gmm, Obs, Ubs, Hel, Ses, Gby, Adc, Rmc, Val, Gac, Cpr, Mrg, Mut, Umt
from ..util.error import EoqError

from copy import deepcopy
from enum import Enum
import re
import traceback

"""
    CONSTANTS & SETTINGS
"""

# Create dictionary from external symbol definition.
# This will be needed later to convert i.e. "!" -> "CLS"
SYMBOLS_2_QRY_DICT = dict((v, k) for k, v in QRY_SYMBOLS.items())

# List all commands that need a list of args and not *args or a mix
QRY_CMDS_W_LIST_INPUT = ["[","&ZIP"]
QRY_CMDS_W_SINGLE_OR_LIST_INPUT = [":","&EQA"]
QRY_CMDS_W_MIXED_INPUT = ["@"]

# Settings of divider symbols
EXPRESSION_DIVIDERS = [";", "\r", "\n"]
QRY_ARG_DIVIDER = ","
CMD_ARG_DIVIDER = " "

# Do not change! This way parsing is needed for re.split
EXPRESSION_DIVIDERS_ = "[" + "".join(EXPRESSION_DIVIDERS) + "]"

# Change this constant to control the context distance of
# the error message if parsing fails.
ERROR_MSG_VIEW_DISTANCE = 3

# Remove unnecessary outer parentheses for py2txt
REMOVE_OUTER_QRY_PARENTHESES = True

# Assign stop symbols to starters
STOPPING_SYMBOL_DICT = {
    "(": ")",
    "{": "}",
    "[": "]",
}

WHITESPACE_REGEX = re.compile("\s")


def DoNothing(*args, **kwargs):
    """ placeholder for unimplemented functions """
    pass


class ParsingCmds(Enum):
    STEP_IN = "INN"
    STEP_OUT = "OUT"


'''
    TEXT Serializer   
'''


class TextSerializer(Serializer):
    def __init__(self):
        """
            TEXTSERIALIZER implements a translation between EOQ expressions and Python commands to
            interact with the EOQ database.
        """

        """
            Toggles verbose output of construction steps within Des()
        """
        self._debugMode = False

        """
            PY2TXT Translators
        """
        self.cmdTranslator = lambda o: o.cmd + " " + self._StripOuterQry(self.Ser(o.a))

        self.qryTranslators = {
            QrySegTypes.OBJ: lambda o: QRY_SYMBOLS[QrySegTypes.OBJ] + self._TranslateQryArgs(o.v),
            QrySegTypes.HIS: lambda o: QRY_SYMBOLS[QrySegTypes.HIS] + self._TranslateQryArgs(o.v),
            QrySegTypes.PTH: lambda o: QRY_SYMBOLS[QrySegTypes.PTH] + self._TranslateQryArgs(o.v),
            QrySegTypes.CLS: lambda o: QRY_SYMBOLS[QrySegTypes.CLS] + self._TranslateQryArgs(o.v),
            QrySegTypes.INO: lambda o: QRY_SYMBOLS[QrySegTypes.INO] + self._TranslateQryArgs(o.v),
            QrySegTypes.MET: lambda o: QRY_SYMBOLS[QrySegTypes.MET] + self._TranslateQryArgs(o.v,True),
            #QrySegTypes.MET: lambda o: QRY_SYMBOLS[QrySegTypes.MET] + "(" + self.Ser(o.v[0]) + (
            #    "," + self.Ser(o.v[1:]) if len(o.v) > 1 else "") + ")",
            QrySegTypes.NOT: lambda o: QRY_SYMBOLS[QrySegTypes.NOT],
            QrySegTypes.CNS: lambda o: QRY_SYMBOLS[QrySegTypes.CNS],
            QrySegTypes.TRM: lambda o: QRY_SYMBOLS[QrySegTypes.TRM] + self._TranslateQryArgs(o.v,True),
            QrySegTypes.TRY: lambda o: QRY_SYMBOLS[QrySegTypes.TRY] + self._TranslateQryArgs(o.v,True),
            QrySegTypes.IDX: lambda o: QRY_SYMBOLS[QrySegTypes.IDX] + self._TranslateQryArgs(o.v,True),
            QrySegTypes.SEL: lambda o: self._TranslateQryArgs(o.v,False,'{','}',True), #"{" + self.Ser(o.v) + "}",
            QrySegTypes.ARR: lambda o: self._TranslateQryArgs(o.v,True,'[',']',True), #"[" + ",".join([self.Ser(a) for a in o.v]) + "]",
            QrySegTypes.ZIP: lambda o: QRY_SYMBOLS[QrySegTypes.ZIP] + self._TranslateQryArgs(o.v,True),
            QrySegTypes.QRY: lambda o: self._TranslateQryArgs(o.v,True,'(',')',True,''), #"(" + "".join([self.Ser(a) for a in o.v]) + ")",
            QrySegTypes.ANY: lambda o: QRY_SYMBOLS[QrySegTypes.ANY] + self._TranslateQryArgs(o.v),
            QrySegTypes.ALL: lambda o: QRY_SYMBOLS[QrySegTypes.ALL] + self._TranslateQryArgs(o.v),
            QrySegTypes.EQU: lambda o: QRY_SYMBOLS[QrySegTypes.EQU] + self._TranslateQryArgs(o.v),
            QrySegTypes.EQA: lambda o: QRY_SYMBOLS[QrySegTypes.EQA] + self._TranslateQryArgs(o.v),
            QrySegTypes.NEQ: lambda o: QRY_SYMBOLS[QrySegTypes.NEQ] + self._TranslateQryArgs(o.v),
            QrySegTypes.LES: lambda o: QRY_SYMBOLS[QrySegTypes.LES] + self._TranslateQryArgs(o.v),
            QrySegTypes.GRE: lambda o: QRY_SYMBOLS[QrySegTypes.GRE] + self._TranslateQryArgs(o.v),
            QrySegTypes.RGX: lambda o: QRY_SYMBOLS[QrySegTypes.RGX] + self._TranslateQryArgs(o.v),
            QrySegTypes.ADD: lambda o: QRY_SYMBOLS[QrySegTypes.ADD] + self._TranslateQryArgs(o.v),
            QrySegTypes.SUB: lambda o: QRY_SYMBOLS[QrySegTypes.SUB] + self._TranslateQryArgs(o.v),
            QrySegTypes.MUL: lambda o: QRY_SYMBOLS[QrySegTypes.MUL] + self._TranslateQryArgs(o.v),
            QrySegTypes.DIV: lambda o: QRY_SYMBOLS[QrySegTypes.DIV] + self._TranslateQryArgs(o.v),
            QrySegTypes.ORR: lambda o: QRY_SYMBOLS[QrySegTypes.ORR] + self._TranslateQryArgs(o.v),
            QrySegTypes.XOR: lambda o: QRY_SYMBOLS[QrySegTypes.XOR] + self._TranslateQryArgs(o.v),
            QrySegTypes.AND: lambda o: QRY_SYMBOLS[QrySegTypes.AND] + self._TranslateQryArgs(o.v),
            QrySegTypes.NAD: lambda o: QRY_SYMBOLS[QrySegTypes.NAD] + self._TranslateQryArgs(o.v),
            QrySegTypes.CSP: lambda o: QRY_SYMBOLS[QrySegTypes.CSP] + self._TranslateQryArgs(o.v),
            QrySegTypes.ITS: lambda o: QRY_SYMBOLS[QrySegTypes.ITS] + self._TranslateQryArgs(o.v),
            QrySegTypes.DIF: lambda o: QRY_SYMBOLS[QrySegTypes.DIF] + self._TranslateQryArgs(o.v),
            QrySegTypes.UNI: lambda o: QRY_SYMBOLS[QrySegTypes.UNI] + self._TranslateQryArgs(o.v),
            QrySegTypes.CON: lambda o: QRY_SYMBOLS[QrySegTypes.CON] + self._TranslateQryArgs(o.v)
        }
        self.priTranslators = {
            # primitive types
            bool: lambda o: str(o),
            int: lambda o: str(o),
            float: lambda o: str(o),
            str: lambda o: self._StringTranslator(o),
            list: lambda o: self._ListTranslator(o),
            type(None): lambda o: "%"
        }

        self.lastTranslation = None

        """
            TXT2PY Constructors
        """
        self.cmdConstructors = {
            # these are only the base constructors
            # chained constructors are handled separately
            # args is a list
            CmdTypes.GET: lambda args: Get(*args),
            CmdTypes.SET: lambda args: Set(*args),
            CmdTypes.ADD: lambda args: Add(*args),
            CmdTypes.REM: lambda args: Rem(*args),
            CmdTypes.MOV: lambda args: Mov(*args),
            CmdTypes.CLO: lambda args: Clo(*args),
            CmdTypes.CRT: lambda args: Crt(*args),
            CmdTypes.CRN: lambda args: Crn(*args),
            CmdTypes.QRF: lambda args: Qrf(*args),
            CmdTypes.CPR: lambda args: Cpr(*args),
            CmdTypes.MRG: lambda args: Mrg(*args),
            CmdTypes.HEL: lambda args: Hel(*args),
            CmdTypes.GBY: lambda args: Gby(*args),
            CmdTypes.SES: lambda args: Ses(*args),
            CmdTypes.GMM: lambda args: Gmm(),
            CmdTypes.RMM: lambda args: Rmm(*args),
            CmdTypes.UMM: lambda args: Umm(*args),
            CmdTypes.STS: lambda args: Sts(),
            CmdTypes.CHG: lambda args: Chg(*args),
            CmdTypes.OBS: lambda args: Obs(*args),
            CmdTypes.UBS: lambda args: Ubs(*args),
            CmdTypes.GAA: lambda args: Gaa(*args),
            CmdTypes.CAL: lambda args: Cal(*args),
            CmdTypes.ASC: lambda args: Asc(*args),
            CmdTypes.ABC: lambda args: Abc(*args),
            CmdTypes.CST: lambda args: DoNothing(),
            CmdTypes.ADC: lambda args: Adc(*args),
            CmdTypes.RMC: lambda args: Rmc(*args if len(args) >= 1 else [None]), # for cmds that can be executed without args
            CmdTypes.VAL: lambda args: Val(*args if len(args) >= 1 else [None]),
            CmdTypes.GAC: lambda args: Gac(),
            CmdTypes.CMP: lambda args: Cmp(),
            CmdTypes.MUT: lambda args: Mut(),
            CmdTypes.UMT: lambda args: Umt()
        }
        self.qryConstructors = {
            # these are only the base constructors
            # chained constructors are handled separately
            # args is a list
            QrySegTypes.OBJ: lambda args: Obj(*args),
            QrySegTypes.HIS: lambda args: His(*args),
            QrySegTypes.PTH: lambda args: Pth(*args),
            QrySegTypes.CLS: lambda args: Cls(*args),
            QrySegTypes.INO: lambda args: Ino(*args),
            QrySegTypes.MET: lambda args: Met(args[0], args[1:]),
            QrySegTypes.NOT: lambda args: Not(),
#             QrySegTypes.TRM: lambda args: Trm(args),
            QrySegTypes.TRY: lambda args: Try(args[0], args[1]),
            QrySegTypes.IDX: lambda args: Idx(args),
            QrySegTypes.ARR: lambda args: Arr(args),
            QrySegTypes.ZIP: lambda args: Zip(args),
            QrySegTypes.QRY: lambda args: args if len(args) else Qry(),
            QrySegTypes.ANY: lambda args: Any(*args),
            QrySegTypes.ALL: lambda args: All(*args),
            QrySegTypes.EQU: lambda args: Equ(*args),
            QrySegTypes.EQA: lambda args: Eqa(args),
            QrySegTypes.NEQ: lambda args: Neq(*args),
            QrySegTypes.LES: lambda args: Les(*args),
            QrySegTypes.GRE: lambda args: Gre(*args),
            QrySegTypes.RGX: lambda args: Rgx(*args)
        }
        self.combinedConstructors = {**self.cmdConstructors, **self.qryConstructors}

        # all segment symbols that are needed for segmentation of the code
        self.cmdAndQryRepresentations = list(QRY_SYMBOLS.values()) + list(CmdTypes.__dict__.values()) + ["}", "]", ")"]

    """ 
        GENERAL METHODS
    """

    def EnableDebugging(self):
        self._debugMode = True

    def DisableDebugging(self):
        self._debugMode = False

    """
        PY 2 TEXT METHODS
    """

    @staticmethod
    def _IsNumerical(string):
        for c in string:
            if not c in ["0","1","2","3","4","5","6","7","8","9","-","+","E","e","."]:
                return False
        return True
    
    @staticmethod
    def _IsNone(string):
        return string == '%'
    
    @staticmethod
    def _ContainsForbiddenCharacter(string):
        # string contains any whitespaces
        if(WHITESPACE_REGEX.search(string)):
            return True
        #conflict with other primitives
        if(string.startswith('+') or 
           string.startswith('-') or 
           string.startswith('.') or 
           string.startswith('%') or 
           string.startswith('0') or 
           string.startswith('1') or 
           string.startswith('2') or 
           string.startswith('3') or 
           string.startswith('4') or 
           string.startswith('5') or 
           string.startswith('6') or 
           string.startswith('7') or 
           string.startswith('8') or
           string.startswith('9') or
           string.lower() == "false" or 
           string.lower() == "true" ):
            return False
        #string contains any query symbol or 
        elif(any(s in string for s in ['#', '$', '/', '!', '?', '@', '&', ':', '{', '[', '(', '=', '~', '<', '>', '^', '\\', '_','|'])):
            return False
        return False
        

    def _StringTranslator(self, string):
        idx = 0
        while idx < len(string):
            if any([self._IsCmdOrQry(x) for x in [string[idx], string[idx:min(idx+3,len(string))], string[idx:min(idx+4,len(string))]]])\
               or self._ContainsForbiddenCharacter(string):
                return "'"+string+"'"
            idx += 1
        return string

    def _ListTranslator(self, li):
        if len(li)>1:
            return "(%s)" % (",".join([self.Ser(x) for x in li]))
        elif len(li) == 1:
            return "(%s,)" % (self.Ser(li[0]))
        else:
            return "()"

    def _StripOuterQry(self, string):
        if REMOVE_OUTER_QRY_PARENTHESES and type(string) == str:
            try:
                # pre-evaluating conditions to avoid out of range string slicing
                cond1 = string[0] == "(" and string[-1] == ")"
            except:
                cond1 = False
            try:
                cond2 = self._IsQry(string[1])
            except:
                cond2 = False
            try:
                cond3 = self._IsQry(string[1:5])
            except:
                cond3 = False
            if cond1 and ( cond2 or cond3 ):
                return string[1:-1]
            else:
                return string
        else:
            return string

    def _TranslateCmd(self, o):
        if o.cmd:
            if o.cmd == CmdTypes.CMP:
                return "\n".join([self.Ser(v) for v in o.a])
            elif type(o.a) == list:
                return o.cmd + " " + " ".join([self._StripOuterQry(self.Ser(v)) if isinstance(v,Qry) else self.Ser(v) for v in o.a ])
            else:
                argStr = self._StripOuterQry(self.Ser(o.a)) if isinstance(o.a,Qry) else self.Ser(o.a)
                if(0 < len(argStr)):
                    return o.cmd + " " + argStr
                else: #commands with no arguments
                    return o.cmd

    def _TranslateQry(self, o):
        if o.qry:
            return self.qryTranslators[o.qry](o)
        
    def _TranslateQryArgs(self,a,multiArgs=False,prefix="(",postfix=")",prePostFixForSingleElement=False,separator=","):
        if(multiArgs and type(a) == list):
            if(prePostFixForSingleElement or len(a)>1):
                return prefix + separator.join([self.Ser(v) for v in a]) + postfix
            elif(len(a)==1):
                return self.Ser(a[0])
            else: #0 element list
                return ''
        else:
            if(prePostFixForSingleElement):
                return prefix + self.Ser(a) + postfix
            else:
                return self.Ser(a)

    def _TranslatePri(self, o):
        return self.priTranslators[type(o)](o)
    
    

    def Ser(self, val):
        """
            Translates a Python command to EOQ code
        """
        try:
            return self._TranslateCmd(val)
        except:
            try:
                return self._TranslateQry(val)
            except:
                try:
                    return self._TranslatePri(val)
                except:
                    raise EoqError(0, "Text serializer failed for %s" % (str(val)))

    """ 
        TEXT 2 PY METHODS 
    """

    @staticmethod
    def _IsCmd(segment):
        if segment in CmdTypes.__dict__.values():
            return True
        else:
            return False

    @staticmethod
    def _IsQry(segment):
        if SYMBOLS_2_QRY_DICT.get(segment, "") in QrySegTypes.__dict__.values():
            return True
        else:
            return False

    def _IsCmdOrQry(self, segment):
        if self._IsCmd(segment) or self._IsQry(segment):
            return True
        else:
            return False

    @staticmethod
    def _IsStarterSymbol(segment):
        return True if segment in ["(", "[", "{"] else False

    @staticmethod
    def _IsStopperSymbol(segment):
        return True if segment in [")", "]", "}"] else False

    @staticmethod
    def _IsQryStart(segment):
        return True if segment == "(" else False

    @staticmethod
    def _IsQryEnd(segment):
        return True if segment == ")" else False

    @staticmethod
    def _ShrinkWhitespace(segments):
        _segments = []
        for seg in segments:
            lastAddedSegment = _segments[-1] if len(_segments) else ""
            if seg == lastAddedSegment and seg == " ":
                pass
            else:
                _segments.append(seg)
        return _segments

    def _UnwrapSingleItemLists(self, ls):
        """
            Flattens single item lists in lists. I. e.
            [1,2,[3]] -> [1,2,3]
            [1,2,[3,4]] -> [1,2,[3,4]]
            This makes the parser more robust to unnecessary query parentheses!
        """
        if type(ls) == list:
            if len(ls) == 1 and (isinstance(ls[0],Qry) or list==type(ls[0])):
                return self._UnwrapSingleItemLists(ls[-1])
            else:
                return [self._UnwrapSingleItemLists(item) for item in ls]
        else:
            return ls

    def _Unwrap(self, ls):
        """
            Avoids flattening of a top level single item list
        """
        ls = self._UnwrapSingleItemLists(ls)
        if type(ls) == list:
            return ls
        else:
            return [ls]

    @staticmethod
    def _IsQryDivider(segment):
        return True if segment == QRY_ARG_DIVIDER else False

    @staticmethod
    def _IsCmdDivider(segment):
        return True if segment == CMD_ARG_DIVIDER else False

    @staticmethod
    def _IsDivider(segment):
        return True if segment in [QRY_ARG_DIVIDER, CMD_ARG_DIVIDER] else False

    def _IsArgument(self, segment):
        if self._IsCmdOrQry(segment) or self._IsDivider(segment) or self._IsStopperSymbol(segment):
            return False
        else:
            return True

    def _GetSegments(self, code):
        """
        GETSEGMENTS splits a textual EOQ expression string into its segments.
        :param code: EOQ expression as a String
        :return: List of segment Strings
        """
        segments = []
        char_buffer = ""
        idx = 0
        while idx < len(code):
            # get current character
            char = code[idx]
            # get words in quotations
            if char in ["\"", "\'"]:
                if char_buffer:
                    segments.append(char_buffer)
                    char_buffer = ""
                n = 1
                while code[idx + n] != char:
                    n += 1
                segments.append("\'"+code[idx + 1:idx + n]+"\'")
                idx += n
            # get single symbol qrys or ws
            elif char in self.cmdAndQryRepresentations or self._IsDivider(char):
                if char_buffer:
                    segments.append(char_buffer)
                    char_buffer = ""
                segments.append(char)
            # get 3-char qrys or cmds
            elif code[idx: idx + 3].upper() in self.cmdAndQryRepresentations:
                if char_buffer:
                    segments.append(char_buffer)
                    char_buffer = ""
                segments.append(code[idx: idx + 3].upper())
                idx += 2
            # get &-codes
            elif char == "&":
                if char_buffer:
                    segments.append(char_buffer)
                    char_buffer = ""
                segments.append(QRY_SYMBOLS[code[idx + 1: idx + 4].upper()])
                idx += 3
            # fill char buffer otherwise
            else:
                char_buffer += char
            idx += 1
        # final buffer append
        if char_buffer:
            segments.append(char_buffer)
        # delete multiple consecutive whitespaces
        return self._ShrinkWhitespace(segments)

    def _GetBoundConstructor(self, cmd, obj):
        """
            returns a bound constructor of obj
        """
        # check if cmd takes list input
        takesList = QRY_SYMBOLS[cmd] in QRY_CMDS_W_LIST_INPUT
        takesSingleOrList = QRY_SYMBOLS[cmd] in QRY_CMDS_W_SINGLE_OR_LIST_INPUT
        takesMix = QRY_SYMBOLS[cmd] in QRY_CMDS_W_MIXED_INPUT

        # convert string to fit method names, i.e. GET -> Get
        cmd = cmd.lower().capitalize()

        # break if non-chainable
        if not hasattr(obj, cmd):
            raise EoqError(0, f"{obj} has no bound method {cmd}!")

        # construct function
        def func(args):
            if takesList:
                return getattr(obj, cmd)(args)
            elif takesMix:
                return getattr(obj, cmd)(args[0], args[1:])
            elif takesSingleOrList:
                return getattr(obj, cmd)(args if len(args)>1 else args[0])
            else:
                return getattr(obj, cmd)(*args)

        return func

    def _GetBaseConstructor(self, cmd):
        """
            returns a base constructor
        """
        key = cmd.upper()
        
        # check if cmd takes list input
        takesList = (key in QRY_SYMBOLS) and (QRY_SYMBOLS[key] in QRY_CMDS_W_LIST_INPUT)
        takesSingleOrList = (cmd in QRY_SYMBOLS) and (QRY_SYMBOLS[key] in QRY_CMDS_W_SINGLE_OR_LIST_INPUT)
        takesMix = (key in QRY_SYMBOLS) and (QRY_SYMBOLS[key] in QRY_CMDS_W_MIXED_INPUT)

        # construct function
        constructor = self.combinedConstructors[key]
        
        def func(args):
            if takesList:
                return constructor(args)
            elif takesMix:
                return constructor(args)
            elif takesSingleOrList:
                return constructor(args if len(args)>1 else args[0])
            else:
                return constructor(args)

        return func
        

    def _ConvertSegmentToFunction(self, seg, res):
        """
            Converts segment symbol to function if possible. I.e. "!" -> Cls()
        """
        if self._IsCmdOrQry(seg):
            if self._IsQry(seg):
                seg = SYMBOLS_2_QRY_DICT[seg]
            cmd = seg
            try:
                func = self._GetBoundConstructor(cmd, res[-1])
                isBoundCmd = True
            except:
                func = self._GetBaseConstructor(cmd)
                isBoundCmd = False
            return func, isBoundCmd
        else:
            raise EoqError(0, f"Segment {seg} is not convertible to a Python Function.")

#     @staticmethod
#     def _BalanceSteps(parsingList):
#         """
#             closes unclosed STEP_IN cmds, (currently unused)
#         """
#         toClose = 0
#         for seg in parsingList:
#             if seg == ParsingCmds.STEP_IN:
#                 toClose += 1
#             elif seg == ParsingCmds.STEP_OUT:
#                 toClose -= 1
#         for closing in range(toClose):
#             parsingList.append(ParsingCmds.STEP_OUT)
#         return parsingList

    def _HandlePrimitives(self, code):
        #find out the right primitive type
        #use the first char to decide
        val = None
        c = code[0]
        if c == '\'': #quoted string
            val = code.strip("\'") #strip quotes
        elif self._IsNone(code):
            val = None
        elif self._IsNumerical(code): #number
            if '.' in code : #float
                val = float(code)
            elif 'E' in code : #engineering float
                val = float(code)
            else: #int
                val = int(code)
        elif code.lower() == 'true': #Boolean True
                val = True
        elif code.lower() == 'false': #Boolean False
            val = False
        else: #unquoted string
            val = code 
        return val

    @staticmethod
    def _StripOuterWhitespace(code):
        return code.strip()

    def _SeparateCodes(self, code):
        codes = re.split(EXPRESSION_DIVIDERS_, code)
        return [self._StripOuterWhitespace(code) for code in codes]

    def _GetParsingList(self, segments):
        """
            Creates a list of segments and parsing commands.
            This function contains the main logic to identify EOQ syntax.
        """
        # list for result
        parsingList = []

        # buffer variables
        lastSeg = None
        nextSeg = None
        cmdDividerSeen = False
        stepInSinceCmdDivider = 0  # flat counter
        stepInSinceQryStart = []  # keeps track of step ins since subqry start, list because nested qrys possible
        expectedStoppingSymbols = []  # keeps track of what stopping symbol is needed to end the current subqry

        # error handling variables
        segmentsCopy = deepcopy(segments)  # necessary for error handling

        while segments:
            seg = segments.pop(0)
            nextSeg = segments[0] if segments else None

            if self._IsStarterSymbol(seg):
                if self._IsArgument(lastSeg):
                    parsingList.append(ParsingCmds.STEP_OUT)
                    if stepInSinceQryStart:
                        stepInSinceQryStart[-1] = max(stepInSinceQryStart[-1] - 1, 0)
                    elif stepInSinceCmdDivider:
                        stepInSinceCmdDivider = max(stepInSinceCmdDivider - 1, 0)
                # keep track of subqueries
                stepInSinceQryStart.append(1)
                expectedStoppingSymbols.append(STOPPING_SYMBOL_DICT[seg])
                # handle difference between argument to subqry and cmd to subqry by shifting the expected stepout
                if self._IsCmdOrQry(lastSeg) and not self._IsStarterSymbol(lastSeg):
                    try:
                        stepInSinceQryStart[-1] += 1
                        stepInSinceQryStart[-2] -= 1
                    except:
                        pass
                # construct parsing list
                parsingList.append(seg)
                parsingList.append(ParsingCmds.STEP_IN)


            elif self._IsCmdOrQry(seg):
                if parsingList \
                        and not self._IsDivider(lastSeg) \
                        and not self._IsStarterSymbol(lastSeg) \
                        and not self._IsStopperSymbol(lastSeg):
                    # STEPOUT RULE for chained cmds and qrys, is disabled on first cmd
                    parsingList.append(ParsingCmds.STEP_OUT)
                    if stepInSinceQryStart:
                        stepInSinceQryStart[-1] = max(stepInSinceQryStart[-1] - 1, 0)
                    elif stepInSinceCmdDivider:
                        stepInSinceCmdDivider = max(stepInSinceCmdDivider - 1, 0)

                if self._IsCmdDivider(lastSeg) and self._IsCmd(seg):
                    # STEPOUT RULE for divided base cmds
                    parsingList.append(ParsingCmds.STEP_OUT)
                    stepInSinceCmdDivider = max(stepInSinceCmdDivider - 1, 0)

                # construct parsing list
                parsingList.append(seg)
                parsingList.append(ParsingCmds.STEP_IN)

                if self._IsCmd(seg):
                    # reset divider flag on new cmd to disable step balancing on first divider after a new cmd
                    cmdDividerSeen = False

                if stepInSinceQryStart:
                    # count step ins within subqrys
                    stepInSinceQryStart[-1] += 1
                elif cmdDividerSeen:
                    # count step ins within separated cmd arg blocks
                    stepInSinceCmdDivider += 1


            elif self._IsCmdDivider(seg):
                if not cmdDividerSeen:
                    cmdDividerSeen = True
                else: 
                    while cmdDividerSeen and stepInSinceCmdDivider:
                        # balancing step outs
                        parsingList.append(ParsingCmds.STEP_OUT)
                        stepInSinceCmdDivider -= 1
                    parsingList.append(seg)


#             elif self._IsQryDivider(seg):
#                 if stepInSinceQryStart:
#                     # reset step-ins since qry to 1 start if qry divider met
#                     stepIns = stepInSinceQryStart.pop()
#                     while (stepIns - 1):
#                         parsingList.append(ParsingCmds.STEP_OUT)
#                         stepIns -= 1
#                     stepInSinceQryStart.append(1)
#                     parsingList.append(seg)


            elif self._IsStopperSymbol(seg):
                if not expectedStoppingSymbols:
                    # found stopping symbol without query start
                    faultyIdx = len(segmentsCopy) - len(segments)
                    errorMsg = f"\nSolving of code failed at segment {faultyIdx}\n" \
                               f"Unexpected stopping symbol {seg}\n" \
                               f"...{segmentsCopy[faultyIdx - ERROR_MSG_VIEW_DISTANCE:faultyIdx]}" \
                               f"-->{segmentsCopy[faultyIdx]}<--" \
                               f"{segmentsCopy[faultyIdx + 1:faultyIdx + 1 + ERROR_MSG_VIEW_DISTANCE]}...\n"
                    raise EoqError(0, errorMsg)
                if seg == expectedStoppingSymbols[-1]:
                    # correct stopping symbol
                    if stepInSinceQryStart:
                        stepIns = stepInSinceQryStart.pop()
                        while (stepIns):
                            parsingList.append(ParsingCmds.STEP_OUT)
                            stepIns -= 1
                    expectedStoppingSymbols.pop()
                else:
                    # wrong stopping symbol
                    faultyIdx = len(segmentsCopy) - len(segments) - 1
                    errorMsg = f"\nSolving of code failed at segment {faultyIdx}\n" \
                               f"Expected {expectedStoppingSymbols[-1]}, got {seg}\n" \
                               f"...{segmentsCopy[faultyIdx - ERROR_MSG_VIEW_DISTANCE:faultyIdx]}" \
                               f"-->{segmentsCopy[faultyIdx]}<--" \
                               f"{segmentsCopy[faultyIdx + 1:faultyIdx + 1 + ERROR_MSG_VIEW_DISTANCE]}...\n"
                    raise EoqError(0, errorMsg)


            else:
                # add argument to parsing list
                parsingList.append(seg)

            lastSeg = seg

        if expectedStoppingSymbols:
            errorMsg = f"\nSolving of code failed.\n" \
                       f"Expected {expectedStoppingSymbols[-1]} but reached end of code.\n"
            raise EoqError(0, errorMsg)

        return parsingList

    def _SolveParsingList(self, parsingList):
        """
            parses segments to actual constructors
        """
        res = []
        cmd = None
        isBoundCmd = None
        seg = None

        if self._debugMode:
            debugCounter = []

        while parsingList:

            lastSeg = seg
            seg = parsingList.pop(0)

            if self._IsCmdOrQry(seg):
                if self._debugMode:
                    # print(f"converting {seg} using res {res}")
                    debugCounter.append(seg)
                if self._IsQryDivider(lastSeg) or self._IsCmdDivider(lastSeg):
                    cmd, isBoundCmd = self._ConvertSegmentToFunction(seg, None)
                else:
                    cmd, isBoundCmd = self._ConvertSegmentToFunction(seg, res)

            elif seg == ParsingCmds.STEP_IN:
                if self._debugMode:
                    if isBoundCmd:
                        print(f"stepping in after bound {debugCounter[-1]}")
                    else:
                        print(f"stepping in after {debugCounter[-1]}")
                res_ = self._SolveParsingList(parsingList)
                res_ = self._Unwrap(res_)
                if res and isBoundCmd:
                    res[-1] = cmd(res_)
                elif res:
                    res.append(cmd(res_))
                else:
                    res = [cmd(res_)]

            elif seg == ParsingCmds.STEP_OUT:
                if self._debugMode:
                    print(f"stepping out returning {res}")
                return res
            elif not self._IsQryDivider(seg) and not self._IsCmdDivider(seg):
                res.append(self._HandlePrimitives(seg))

        if self._debugMode:
            print(f"stepping out with args {res}")
        return res

    def _ConstructCompound(self, commandList):
        """
            combines multiple commands to a compound
        """
        res = Cmp()
        for idx, cmd in enumerate(commandList):
            try:
                res.a.append(cmd)
#                 # get bound cmd for cmp
#                 func = getattr(res, cmd.cmd.lower().capitalize())
#                 # get content
#                 args = cmd.a
#                 if type(args) != list:
#                     args = [args]
#                 # construct
#                 res = func(*args)
            except Exception as e:
                errorMsg = f"\nSolving of code failed.\n" \
                           f"Could not add command {idx + 1} to Compound.\n"
                raise EoqError(0, errorMsg)
        return res

    def Des(self, code):
        """
            DES takes a textual representation EOQ code and constructs the cmd/qry chain
        """
        results = []
        codes = self._SeparateCodes(code)
        for code in codes:
            # split code into segments
            segments = self._GetSegments(code)
            # inject parsing commands
            parsingList = self._GetParsingList(segments)
            if self._debugMode:
                print(parsingList)
            # deepcopy parsingList for error handling
            parsingListCopy = deepcopy(parsingList)
            # solve combined segments and parsing commands
            try:
                command = self._SolveParsingList(parsingList)
                results += command
            except:
                # solver failed
                traceback.print_exc()
                faultyIdx = len(parsingListCopy) - max(1, len(parsingList))
                errorMsg = f"\nSolving of code failed at segment {faultyIdx}\n" \
                           f"...{(parsingListCopy[faultyIdx - ERROR_MSG_VIEW_DISTANCE:faultyIdx])}" \
                           f" -->{parsingListCopy[faultyIdx]}<-- " \
                           f"{(parsingListCopy[faultyIdx + 1:faultyIdx + 1 + ERROR_MSG_VIEW_DISTANCE])}...\n"
                raise EoqError(0, errorMsg)

        if len(results) == 1:
            return results[0]
        else:
            return self._ConstructCompound(results)

"""
    Quick overview for future maintanence.
    Des() calls the following functions in sequence:
    - _SeparateCodes() to detect if code is a compound of multiple codes
    - _GetSegments() to split the code into a list of segments
    - _GetParsingList()** identifies the EOQ syntax and constructs a modified list for recursive solving
    - _SolveParsingList() solves said list into a Python command
    - _ConstructCompound() if multiple codes were processed
    
    ** This function is the most difficult to debug, stepping the execution while considering the wanted behaviour at
    helps a lot.
"""


"""Definition of meta model 'constraintmodel'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'constraintmodel'
nsURI = 'http://www.example.org/constraintmodel'
nsPrefix = 'constraintmodel'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)
ConstraintTypeEnum = EEnum('ConstraintTypeEnum', literals=['SOFT', 'HARD'])


class Constraint(EObject, metaclass=MetaEClass):

    constraintType = EAttribute(eType=ConstraintTypeEnum, derived=False, changeable=True)
    target = EAttribute(eType=EString, derived=False, changeable=True)
    feature = EAttribute(eType=EString, derived=False, changeable=True)
    law = EAttribute(eType=EString, derived=False, changeable=True)
    name = EAttribute(eType=EString, derived=False, changeable=True)
    annotation = EAttribute(eType=EString, derived=False, changeable=True)

    def __init__(self, *, constraintType=None, target=None, feature=None, law=None, name=None, annotation=None, **kwargs):
        if kwargs:
            raise AttributeError('unexpected arguments: {}'.format(kwargs))

        super().__init__()

        if constraintType is not None:
            self.constraintType = constraintType

        if target is not None:
            self.target = target

        if feature is not None:
            self.feature = feature

        if law is not None:
            self.law = law

        if name is not None:
            self.name = name

        if annotation is not None:
            self.annotation = annotation



import asyncio
import threading
import tornado.ioloop
import tornado.web
import tornado.websocket
import signal
import time

from uuid import uuid4

import traceback

import argparse

from eoq2 import __version__ as eoqVersion

from eoq2.mdb.pyecore import PyEcoreWorkspaceMdbProvider,PyEcoreMdbAccessor,PyEcoreIdCodec
from eoq2.domain.local import LocalMdbDomain

from eoq2.action.externalpy import ExternalPyScriptHandler
from eoq2.serialization import JsonSerializer,JsSerializer
from eoq2.util import ConsoleAndFileLogger,Backupper,LogLevels

from timeit import default_timer as timer

from eoq2.frame import DomainFrameHandler,MultiVersionFrameHandler,Frame,FrameTypes
from eoq2.legacy import LegacyDomain,LegacyTextFrameHandler

sessions = {} #store all sessions to prevent to create a dublicated session id

def CreateNewUniqueSessionId():
    global sessions
    sessionId = str(uuid4())
    while(sessionId in sessions):
        sessionId = str(uuid4())
    sessions[sessionId] = sessionId
    return sessionId

class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write("Sorry, this is the wrong way to access the PyEoq-Web-Server.\nClone https://gitlab.com/mbrunner/ecoreeditor and start the index.html.")
        
class EoqHandler(tornado.web.RequestHandler):
    def initialize(self,eoq,serializer):
        self.start = timer()
        self.eoq = eoq
        self.serializer = serializer
        
    def post(self):
        cmdJson = self.get_body_argument("command")
        cmd = self.serializer.deserialize(cmdJson)
        result = self.eoq.Do(cmd)
        response = self.serializer.serialize(result)
        #output the response
        self.set_header("Content-Type", "text/plain")
        self.set_header('Access-Control-Allow-Origin','*') #Necessary to allow local requests
        self.write(response)
        self.end = timer()
        print("Request completed in %s s."%(self.end-self.start))
        
class PyEoq2WebSocket(tornado.websocket.WebSocketHandler):
    def initialize(self,frameHandler,serializer,logger):
        self.sessionId = CreateNewUniqueSessionId()
        self.frameHandler = frameHandler
        self.serializer = serializer
        self.logger = logger 
        #listen to events
        self.frameHandler.Observe(self.OnEvents,context=self,sessionId=self.sessionId) #only observe events specific to this session
        self.eventQueue = [] #the list of events to be send when it is next possible
        self.eventQueueMutex = asyncio.Lock()
        self.eventQueueMutex2 = threading.Lock();
        self.ioLoop = tornado.ioloop.IOLoop.current()
        
    #the check_origin override allows request from the local server
    def check_origin(self, origin):
        return True
    
    def open(self):
        pass

    def on_message(self, message):
        #print("WebSocket received: "+ message)
        start = timer()
        self.logger.PassivatableLog(LogLevels.DEBUG,lambda : "Web socket message received: %s"%(message))
        try:
            frames = self.serializer.Des(message)
            frames= self.frameHandler.Handle(frames,sessionId=self.sessionId)
            response = self.serializer.Ser(frames)
            self.write_message(response)
            end = timer()
            self.logger.PassivatableLog(LogLevels.DEBUG,lambda : "Web socket response (after %d s): %s"%(end-start,response))
        except Exception as e:
            self.logger.Error("Invalid message received: %s (error: %s)"%(message,str(e)))
            self.logger.Error(traceback.format_exc())
            #raise Exception()

    def on_close(self):
        #quit listening to events
        self.frameHandler.Unobserve(self.OnEvents,context=self)
        #force closing the session
        self.frameHandler.Gby(self.sessionId)
        
    def SendEvents(self):
        # !! This must be called in the main thread of tornado !!
        self.eventQueueMutex2.acquire()
        try:
        #with self.eventQueueMutex:
            # pack events in frames    
            frames = []
            for evt in self.eventQueue:
                frame = Frame(FrameTypes.EVT,0,evt)
                frames.append(frame)
            # clear the event queue
            self.eventQueue.clear()
            # send events 
            response = self.serializer.Ser(frames)
            if(isinstance(response, str)):
                self.write_message(response) 
            else:
                self.logger.Warn("Got non-string response of type %s: %s"%(type(response).__name__,str(response)))
        finally:
            self.eventQueueMutex2.release();
    
    def OnEvents(self,evts,src):
        #!! Attention this is called from the outside (from any thread) and is, therefore, 
        # not in the tornado IO loop and can not safely call write!!!
        self.eventQueueMutex2.acquire();
        try:
        #with self.eventQueueMutex:
            for evt in evts:
                self.eventQueue.append(evt)
            # make sure the events are processed in the main thread
            self.ioLoop.add_callback(self.SendEvents)
        finally:
            self.eventQueueMutex2.release();
        

class PyeoqWebSocketServer:
    def __init__(self):
        self.ioLoop = None
        self.backupper = None
        self.logger = None
        self.mdbProvider = None
        self.valueCodec = None
        self.mdbAccessor = None
        self.domain = None
        self.frameHandler = None
        self.externalActionHandler = None
        self.app = None
        self.httpServerApi = None
        self.args = None
        self.isRunning = False
        

    def Start(self,args):
        self.args = args
        self.isRunning = True
        #start local ioloop
        self.ioLoop = tornado.ioloop.IOLoop()
        self.ioLoop.make_current()
        
        
        #Do backup if required
        if(args.backup):
            self.backupper = Backupper([args.workspaceDir,args.logDir])
            self.backupper.CreateBackup()
        
        #initialize logger. For the file based logger this must happen after the backup
        #logger = ConsoleLogger()
        self.logger = ConsoleAndFileLogger(logDir=args.logDir,toConsole=args.logToConsole,toFile=args.logToFile,activeLevels=[LogLevels.INFO,LogLevels.WARN,LogLevels.ERROR,"change","event"])
        #define a global serializer for outputs
        self.debugSerializer = JsSerializer()
              
        #initialize EOQ
        self.mdbProvider = PyEcoreWorkspaceMdbProvider(args.workspaceDir,metaDir=[args.metaDir],saveTimeout=args.autosave,logger=self.logger,trackFileChanges=args.trackFileChanges)
        self.valueCodec = PyEcoreIdCodec()
        self.mdbAccessor = PyEcoreMdbAccessor(self.mdbProvider.GetMdb(),self.valueCodec)
        self.domain = LocalMdbDomain(self.mdbAccessor,maxChanges=args.maxChanges,logger=self.logger,serializer=self.debugSerializer, enableBenchmark=args.enableBenchmark)
        self.mdbProvider.CoupleWithDomain(self.domain, self.valueCodec)
        
        #register external actions if required
        if(args.actions):
            self.externalActionHandler = ExternalPyScriptHandler(self.domain.cmdRunner.callManager,args.actionsDir,logger=self.logger)
        
        
        
        #initialize frame handlers in order to support legacy frames
        self.frameHandler = MultiVersionFrameHandler([
            (100,LegacyTextFrameHandler(LegacyDomain(self.domain,serializer=self.debugSerializer,logger=self.logger),logger=self.logger)),
            (200,DomainFrameHandler(self.domain))
            ])
        
        #define serializer for all client communication
        self.serializer = JsonSerializer()
        
        #start webserver
        print("Initiating tornado application... ",end='')
        self.app = tornado.web.Application([
            (r"/", MainHandler),
            (args.ws, PyEoq2WebSocket, dict(frameHandler=self.frameHandler,serializer=self.serializer,logger=self.logger))
        ])
        self.httpServerApi = tornado.httpserver.HTTPServer(self.app)
        print("ok")
        
        port = args.port
        print("Opening port %d... "%(port),end='')
        self.httpServerApi.listen(port)
        print("ok")
        
        print("Server is running...")
        tornado.ioloop.IOLoop.current().start()
        
    def Stop(self,sig,frame):
        print('Shutting down...')
        print('Stopping tornado... ',end='')
        asyncio.run(self.httpServerApi.close_all_connections()) # Is this necessary? Seems not possible to wait here
        self.httpServerApi.stop()
        self.ioLoop.stop()
        time.sleep(3) #wait for tornado to close
        print('ok')
        print('Closing the domain... ',end='')
        self.domain.Close()
        print('ok')
        print('Closing the mdb... ',end='')
        self.mdbProvider.Close()
        print('ok')
        self.isRunning = False



if __name__ == "__main__":
    version = "2.2.1" #to be changed manually in accordance to EOQ version
    #get commandline arguments
    parser = argparse.ArgumentParser(description='An eoq2 server listening for commands on a web socket.')
    parser.add_argument('--ws', metavar='ws', type=str, default='/ws/eoq.do', help='the websocket url', dest='ws')
    parser.add_argument('--port', metavar='port', type=int, default=8000, help='the port name the EOQ server shall listen to', dest='port')
    parser.add_argument('--workspaceDir', metavar='workspaceDir', type=str, default='./workspace', help='the root directory including the model files to work with', dest='workspaceDir')
    parser.add_argument('--metaDir', metavar='metaDir', type=str, default='./.meta', help='The directory including the meta models of known model formats. This is relative to the workspace', dest='metaDir')
    parser.add_argument('--actions', metavar='actions', type=int, default=1, help='Whether external actions shall be available or not. (0=no, 1=yes)', dest='actions')
    parser.add_argument('--actionsDir', metavar='actionsDir', type=str, default='./workspace/actions', help='The directory containing action files', dest='actionsDir')
    parser.add_argument('--timeout', metavar='timeout', type=float, default=10.0, help='The maximum timeout for a transaction', dest='timeout')  
    parser.add_argument('--backup', metavar='backup', type=int, default=1, help='Shall a backup of the workspace and logs be created? (0=no, 1=yes)', dest='backup')   
    parser.add_argument('--backupDir', metavar='backupDir', type=str, default='./backup', help='Destination folder for the backup', dest='backupDir')   
    parser.add_argument('--logDir', metavar='logDir', type=str, default='./log', help='Destination folder for log files', dest='logDir')  
    parser.add_argument('--logToConsole', metavar='logToConsole', type=int, default=0, help='Print log messages in the console? (0=no, 1=yes)', dest='logToConsole')   
    parser.add_argument('--logToFile', metavar='logToFile', type=int, default=1, help='Print log messages in log files? (0=no, 1=yes)', dest='logToFile')   
    parser.add_argument('--autosave', metavar='autosave', type=float, default=5.0, help='The timeout after that autosave is triggered', dest='autosave')   
    parser.add_argument('--trackFileChanges', metavar='trackFileChanges', type=int, default=1, help='Update models if files in the workspace dir change? (0=no, 1=yes)', dest='trackFileChanges')
    parser.add_argument('--maxChanges', metavar='maxChanges', type=int, default=10000, help='How many changes shall be remembered until the oldest change is forgotten (1 to 1000000)', dest='maxChanges') 
    parser.add_argument('--enableBenchmark', metavar='enableBenchmark', type=int, default=0, help='Messure the time for queries and commands and write log when closed? (0=no, 1=yes)', dest='enableBenchmark')     
    args = parser.parse_args()

    print("*******************************************")
    print("*           EOQ2  Web Server              *")
    print("*******************************************")
    print("Version:      %s"%(version))
    print("EOQ Version:  %s"%(eoqVersion))
    print("Workspace:    %s"%(args.workspaceDir))
    print("Web Socket:   %s"%(args.ws))
    print("Port:         %s"%(args.port))
    print("Max Changes:  %d"%(args.maxChanges))
    print("Track files:  %d"%(args.trackFileChanges))
    print("Benchmark:    %d"%(args.enableBenchmark))
    print("*******************************************")
    
    server = PyeoqWebSocketServer()
    serverThread = threading.Thread(target=server.Start, args=(args,))
        
    print("Registering signal listeners ... ",end='')
    signal.signal(signal.SIGTERM, server.Stop)
    signal.signal(signal.SIGINT,  server.Stop)
    print("ok")
    
    serverThread.start()
    time.sleep(1) #wait for the server thread to start
    
    #keep the main thread alive until the server is closed. This is mandatory to receive CTRL + C signals
    while server.isRunning:
        time.sleep(3)
        
    print('Goodbye!')
    


'''
 2019 Bjoern Annighoefer
'''

'''
    Server
'''

from ..domain import Domain
from ...serialization import JsonSerializer
from ...frame import FrameTypes,Frame
from ...event import EvtTypes

from multiprocessing import Queue
import queue # imported for using queue.Empty exception
   
import threading
import traceback

class MultiprocessingQueueDomainHost(Domain):
    def __init__(self,domain,callManager,serializer=JsonSerializer(),timeout=0.1,sessionId=None,callId=None):
        super().__init__()
        self.domain = domain
        self.callManager = callManager
        self.cmdQueue = Queue()
        self.resQueue = Queue()
        self.evtQueue = Queue()
        self.timeout = timeout
        self.sessionId = sessionId
        self.callId = callId
        
        self.serializer = serializer
        self.shallRun = True
        
        #listen to domain events
        self.domain.Observe(self.OnDomainEvent,context=self,sessionId=self.sessionId)
        
    def Start(self):
        self.listenerThread = threading.Thread(target=self.__RemoteCommandHandler, args=())
        self.listenerThread.start()
        
    def Join(self):
        self.domain.Unobserve(self.OnDomainEvent,context=self)
        self.shallRun = False
        self.listenerThread.join()
        
    def __RemoteCommandHandler(self):
        while self.shallRun:
            try:
                frameStr = self.cmdQueue.get(timeout=self.timeout)
                frames = self.serializer.Des(frameStr)
                outFrames = []
                for frame in frames:
                    if(frame.eoq==FrameTypes.CMD):
                        cmd = frame.dat
                        res = self.domain.RawDo(cmd,sessionId=self.sessionId)
                        frame.eoq = FrameTypes.RES
                        frame.dat = res
                        outFrames.append(frame)
                    elif(frame.eoq == FrameTypes.EVT):
                        evt = frame.dat
                        if(evt.evt == EvtTypes.OUP):
                            self.callManager.AddCallOutputRaw(evt.a[0],evt.a[1],evt.a[2])
                        elif(evt.evt == EvtTypes.CST):
                            callId = evt.a[0]
                            status = evt.a[1]
                            info = evt.a[2]
                            self.callManager.ChangeCallStatus(callId,status,info)
                        elif(evt.evt == EvtTypes.CVA):
                            self.callManager.SetCallValue(evt.a[0],evt.a[1])
                #send a cumulative answer
                if(len(outFrames)>0):
                    frameStr = self.serializer.Ser(outFrames)
                    self.resQueue.put(frameStr) 
                    
            except queue.Empty:
                continue
            except Exception as e:
                print("Error in MultiprocessingQueueDomainHost: %s"%(str(e)))
                traceback.print_exc()
        #thread ends here
            
    def OnDomainEvent(self,evts,source):
        outFrames = []
        for evt in evts:
            if(evt.evt in [EvtTypes.OUP,EvtTypes.CST,EvtTypes.CVA] and self.callId==evt.a[0]): #do not feedback own events
                continue
            outFrames.append(Frame(FrameTypes.EVT,0,evt))
        # if there are events, than forward them
        if(len(outFrames)>0):
            frameStr = self.serializer.Ser(outFrames)
            self.evtQueue.put(frameStr)


from ..mdb import DomainA,CommandA
from pyeoq import ResultParser,CommandParser


from urllib.parse import urlencode
from http.client import HTTPConnection

class RemoteHttpDomain(DomainA):
    
    connection = None
    timeout = 1
    
    def __init__(self, host, port=8000, timeout=1, **kwargs):
        
        self.connection = HTTPConnection(host, port=port)
        
        self.timeout = timeout
        super().__init__(**kwargs)
        self.name = 'RemoteHTTPDomain'
        
        
        
        
    def Do(self, command):
            
        if isinstance(command,CommandA):
            command = CommandParser.CommandToString(command)
        elif not type(command) == str:
            raise TypeError("Argument must be either a string or an instance of CommandA.")
            
        commandStr = urlencode({'command':command})
        
        res = self._post_request(commandStr)
        
        return res
    
    def _post_request(self,commandStr):
        
        self.connection.request("POST", "/eoq.do", commandStr)
        response = self.connection.getresponse()
        resultStr = response.read().decode('utf-8')
        res = ResultParser.StringToResult(resultStr)
        return res


"""
 Workspace MDB Provider
 Bjoern Annighoefer 2019
"""
# todo: catch permission errors when renaming a folder in which is a file opened or the folder is open in the explorer!!!
# todo: the web surface does not recognize when to folders are named the same name -> on explorer side it is recognized
# todo: better printout when something happens -> make paths better readable etc
# todo: the funktionszuweisung part of the event handler should be reworked to get it nicer to read
# todo: there is a known issue that when a newly created file is deleted before it is loaded, it gets recreated again
#       after it has been deleted. Recreate issue: via Web surface create .oaam file and immediately delete it in the explorer

from ...util import NoLogging
from ...event import MsgEvt,ChgTypes,EvtTypes
from .pyecoremdb import PyEcoreMdb

from .workspacemdbmodel import Workspace,ModelResource,Directory,FileResourceTypesE,PathElementA,FileResourceA,XmlResource
from .xmlresourcemodel import Document,Element,Attribute

from pyecore.resources import ResourceSet, URI
from pyecore.ecore import EPackage,EObject


from eoq2 import Cmp, Set, Add
import os
import platform
import shutil
import glob
import time
import xml.etree.ElementTree
import xml.dom.minidom
from pathlib import Path
from timeit import default_timer
from threading import Timer
import traceback

# try importing for watchdog functions
try:
    from watchdog.observers import Observer
    from watchdog.events import PatternMatchingEventHandler, FileSystemEvent, FileCreatedEvent
except ImportError:
    raise ImportError("Could not import Element from watchdog.\nPlease install the watchdog library.")


import pyecore.behavior as behavior  # We need to import the 'behavior' package

# Ugly global necessary for the path methods 
PROVIDER_INSTANCE = None


#Implement methods
@PathElementA.behavior
def actualPath(self):
    path = None
    if(PROVIDER_INSTANCE):
        path = PROVIDER_INSTANCE.GetElementPath(self)
    return path

@PathElementA.behavior
def actualPathAbs(self):
    path = None
    if(PROVIDER_INSTANCE):
        path = PROVIDER_INSTANCE.GetAbsElementPath(self)
    return path

@PathElementA.behavior
def actualPathCwd(self):
    path = None
    if(PROVIDER_INSTANCE):
        path = PROVIDER_INSTANCE.GetElementPath(self)
    return path

# intended for saving changes for later, but file system changes must be checked immediately in order to prevent unwanted situations
# class WorkspaceChgTypes:
#     RESOURCE_ADD = "RESOURCE_ADD"
#     RESOURCE_MOV = "RESOURCE_MOV"
#     RESOURCE_DEL = "RESOURCE_DEL"
#     DIRECTORY_ADD = "DIRECTORY_ADD"
#     DIRECTORY_MOV = "DIRECTORY_MOV"
#     DIRECTORY_DEL = "DIRECTORY_DEL"
        
        
class PyEcoreWorkspaceMdbProvider():
    def __init__(self,baseDir,metaDir=['./.meta'],saveTimeout=10.0,logger=NoLogging(),autoload=True,trackFileChanges=False):
        super().__init__()

        self.baseDir = os.path.normpath(baseDir)
        self.baseDirAbs = os.path.join(os.getcwd(),self.baseDir)
        self.baseDirUri = URI(self.baseDir) 
        self.rset = ResourceSet()
        self.modelResourceLut = {}  #a look-up table between eResources and ModelResource objects
        self.eResourceLut = {}  #a look-up table between ModelResource and eResources objects
        self.eResourcesConnected = False
        self.metaDir = metaDir  #the directory that contains model definitions
        
        self.logger = logger
        
        self.modelroot = Workspace(name='.')  #clone behavior of the old localdomain
        self.lastPersistentPaths = {}  # dir or resources -> path
        self.dirtyResources = []  #stores all modified resources, such that they can be saved
        self.deletedResources = {}  #stores all resources marked for deletion
        self.dirtyObjects = []  #stores all objects that have been modified in order to do not mark them twice
        self.saveTimeout = saveTimeout  #This time is waited for any model modifications and than autosafe is called
        self.saveTimer = None
        self.trackFileChanges = trackFileChanges
        
        self.knownModelExtensions = ['ecore']  #this is used to identify the model files to be loaded. ecore files are known as model files by default.
        self.knownXmlExtensions = ['xml']
        
        self.valueCodec = None  # is set during coupling
        self.domain = None  # is set during coupling
        self.mdb = None  # is created later

        self.exchangeChangeEventsList = []  # this list exchanges events between the eoq and the watchdog to prevent them to trigger each other it has the
                                            # structure that the event is given in a tuple: (string_to_identify_event_type, path_of_changed_element)

        self.modifiedEventSkipList = []
        self.__observer = None
        if(self.trackFileChanges):
            self.logger.Info("File change tracking enabled.")
            self.__observer = Observer()  # create observer for watchdog library
        
        if(autoload):
            self.Load()
            
        #register this as an singelton
        global PROVIDER_INSTANCE
        PROVIDER_INSTANCE = self
        
        
    def Close(self):
        #make sure any changes are saved
        #self.logger.Info("Closing workspace MDB ...") #makes problems when main thread has ended already

        self.__StopDelayedSaving()
        self.__SaveAllDirtyResources()

        #todo: there is a problem with the del which has to be investigated
        if(self.trackFileChanges):
            self.__observer.stop()
            self.__observer.join(10)  # wait until the watchdog observing thread terminates or exceeded timeout
            
        self.logger.Info("ok")
        
    def GetMdb(self):
        return self.mdb
    
    def Load(self):
        self.logger.Info("Loading workspace MDB for %s (meta=%s)..."%(self.baseDir,self.metaDir))
        start = default_timer()
        self.__LoadMetaModels()
        self.__LoadResourceTree()
        end = default_timer()
        self.mdb = PyEcoreMdb(self.modelroot,self.rset.metamodel_registry,self)
        self.logger.Info("Workspace MDB ready after %f s"%(end-start))
    
    def CoupleWithDomain(self,domain,valueCodec):
        self.domain = domain
        self.valueCodec = valueCodec
        self.domain.Observe(self.OnChange,[EvtTypes.CHG])

        # create the observer to track changes made on the workspace
        if(self.trackFileChanges):
            my_event_handler = self.__CreateCustomEventHandlerForWatchdog()  # create custom event handler which is imported from separate file
            self.__observer.schedule(my_event_handler, self.baseDir, recursive=True)  # set up observer
            self.__observer.start()  # start watchdog to observe workspace
            self.logger.Info("Watchdog initialized and ready!")
        
    ''' sync model changes with workspace '''
        
    def OnChange(self,evts,src):
        for evt in evts:

            chg = evt
            ctype = chg.a[1]
            target = self.valueCodec.Dec(chg.a[2]) #reverse what the domain did to the element
            feature = chg.a[3]
            oldOwner = self.valueCodec.Dec(chg.a[6])
            
            if(isinstance(target,FileResourceA)):
                if(ctype==ChgTypes.SET and "name"==feature):
                    # get name and check whether the renaming of the resource was made by the watchdog
                    newName = chg.a[4]
                    # check if actual path is not none so the renaming of a newly created object by creation in the ecoreeditor which
                    # obviously has no actual path throws no error
                    if target.actualPath() is not None and self.__IsEventInEventsExchangeList(Path(self.baseDir,target.actualPathAbs()),"rename"):
                        self.__RemoveEventFromEventsExchangeList(Path(self.baseDir,target.actualPath()),newName)
                        self.logger.Info(f"Renaming Resource to '{newName}' made by the Watchdog recognized!")
                    else:
                        # rename action
                        self.__RenameResource(target,newName)
                else: #handle any other resource change like an object change
                    self.__SetObjectDirty(target)
                    self.__SetObjectDirty(oldOwner)
            elif(isinstance(target,Directory)):
                # directories might be renamed or deleted or new resources/subdirs added or deleted
                if(ctype==ChgTypes.SET and "name"==feature ):
                    # resource renamed
                    newName = chg.a[4]
                    # check whether the event caused by a change, the watchdog made on the eoq
                    # evaluate if the actual path of the target is not None to prevent an error when a newly created object, which is not added
                    # to the eoq, is renamed.
                    if target.actualPath() is not None and self.__IsEventInEventsExchangeList(Path(self.baseDir,target.actualPath()),"rename"):
                        self.__RemoveEventFromEventsExchangeList(Path(self.baseDir,target.actualPath()),"rename")
                        self.logger.Info(f"Renaming directory to '{newName}' made by the Watchdog recognized!")
                    else:
                        self.__RenameDirectory(target,newName)
                elif(ctype==ChgTypes.ADD and "resources"==feature):
                    # new resource
                    resource = self.valueCodec.Dec(chg.a[4])
                    # check if change was made by watchdog
                    if self.__IsEventInEventsExchangeList(Path(self.baseDir,target.actualPath(),resource.name),"create"):
                        self.__RemoveEventFromEventsExchangeList(Path(self.baseDir,target.actualPath(),resource.name),"create")
                        self.logger.Info(f"Creation of file '{resource.name}' made by the Watchdog recognized!")
                    else:
                        if(oldOwner):
                            self.__MoveResource(resource)
                        else:
                            # make entry in events exchange list so the watchdog knows that the change was made by the eoq
                            self.__AddEventToEventsExchangeList(Path(self.baseDir,target.actualPath(),resource.name),"create")
                            self.__AddResource(target,resource)
                elif(ctype==ChgTypes.REM and "resources"==feature):
                    # resource deleted
                    resource = self.valueCodec.Dec(chg.a[4])
                    # check if change was made by the watchdog
                    if self.__IsEventInEventsExchangeList(Path(self.baseDir,target.actualPath(),resource.name),"delete"):
                        self.__RemoveEventFromEventsExchangeList(Path(self.baseDir,target.actualPath(),resource.name),"delete")
                        self.logger.Info(f"Deletion of file '{resource.name}' made by the Watchdog recognized!")
                    else:
                        self.__DeleteResource(resource)
                elif(ctype==ChgTypes.ADD and "subdirectories"==feature):
                    # new directory
                    directory = self.valueCodec.Dec(chg.a[4])
                    if(oldOwner):
                        self.__MoveDirectory(directory)
                    else:
                        # check whether the event is because of a change from the watchdog
                        if self.__IsEventInEventsExchangeList(Path(self.baseDir,directory.actualPath()),"create"):
                            self.__RemoveEventFromEventsExchangeList(Path(self.baseDir,directory.actualPath()),"create")
                            self.logger.Info(f"Creation of directory '{directory.name}' made by the Watchdog recognized!")
                        else:
                            self.__AddDirectory(target,directory)
                elif(ctype==ChgTypes.REM and "subdirectories"==feature):
                    # directory deleted
                    directory = self.valueCodec.Dec(chg.a[4])
                    # get directory of deleted directory relative to the base dir and make it
                    # relative to the base dir with the naming convention of the eoq
                    absolutePathDeletedDir = self.__GetLastPersitentPath(directory)
                    # check if dir deletion is made by the watchdog
                    if self.__IsEventInEventsExchangeList(Path(absolutePathDeletedDir),"delete"):
                        self.__RemoveEventFromEventsExchangeList(Path(absolutePathDeletedDir),"delete")
                        self.logger.Info(f"Deletion of directory '{directory.name}' made by the Watchdog recognized!")
                    else:
                        self.__DeleteDirectory(directory)
            elif(isinstance(target,EObject)):
                self.__SetObjectDirty(target)
                self.__SetObjectDirty(oldOwner)
            elif(ctype==ChgTypes.ADC or ctype==ChgTypes.RMC):
                index = None
                for i in range(len(self.mdb.Root().eContents)):
                    if self.mdb.Root().eContents[i].name == 'cdb.constraintmodel':
                        index = i
                if not(index == None):
                    self.__SetObjectDirty(self.mdb.Root().eContents[index].eContents[0])
                
    ''' resource sync functions '''
       
    def __AddResource(self,dir,resource):
        if(dir): #otherwise there is no need to move the resource
            try:
                #move the file 
                newPath = self.GetAbsElementPath(resource)
                if(newPath): #if this is false, then the resource was added to a directory not attached to the workspace
                    #create placeholder file
                    open(newPath, 'w').close()
                    self.__SetLastPersistentPath(resource, newPath)
                    self.__SetResourceDirty(resource)
                    #update the pyecore URI
                    if(resource.type == FileResourceTypesE.MODEL):
                        eResource = self.rset.create_resource(newPath)
                        self.__LinkModelResourceAndEResource(resource,eResource)
                    #inform about action
                    newRelPath = os.path.relpath(newPath,self.baseDir)
                    self.logger.Info("Added resource %s."%(newRelPath))
            except Exception as e:
                self.logger.Error("Failed to add resource: %s"%(str(e)))
                traceback.print_exc()            
                   
    def __RenameResource(self,resource,newName):
        lastPersistentPath = self.__GetLastPersitentPath(resource)
        if(lastPersistentPath): #otherwise this is a new resource and needs no renaming
            oldPath = lastPersistentPath
            filePath, oldName = os.path.split(oldPath)
            if(newName!=oldName):
                try:
                    # create entry in the events exchange list to tell watchdog that the PyEoq created the file
                    self.__AddEventToEventsExchangeList(Path(filePath,newName), "rename")
                    # move the file
                    newPath = os.path.join(filePath,newName)
                    os.rename(oldPath,newPath) 
                    self.__RefreshResourceOrDirectoryPath(resource)
                    # inform about action
                    oldRelPath = os.path.relpath(oldPath,self.baseDir)
                    newRelPath = os.path.relpath(newPath,self.baseDir)
                    self.logger.Info("Renamed resource %s to %s."%(oldRelPath,newRelPath))
                except FileExistsError:
                    # remove entry from the exchange events list if the renaming could not be made
                    if self.__IsEventInEventsExchangeList(Path(filePath,newName), "rename"):
                        self.__RemoveEventFromEventsExchangeList(Path(filePath,newName), "rename")
                    self.logger.Error(f"Failed to rename resource '{oldName}' to '{newName}.\n"
                                      f"A file with this name already exists in '{filePath}'.")
                except Exception as e:
                    # remove entry from the exchange events list if the renaming could not be made
                    if self.__IsEventInEventsExchangeList(Path(filePath,newName), "rename"):
                        self.__RemoveEventFromEventsExchangeList(Path(filePath,newName), "rename")
                    self.logger.Error("Failed to rename resource: %s"%(str(e)))
                    traceback.print_exc()
                
    def __MoveResource(self,resource):
        lastPersistentPath = self.__GetLastPersitentPath(resource)
        if(lastPersistentPath): #otherwise there is no need to move the resource
            oldPath = lastPersistentPath
            newPath = self.GetAbsElementPath(resource)
            if(newPath!=oldPath):
                try:
                    #move the file 
                    os.rename(oldPath,newPath) 
                    self.__RefreshResourceOrDirectoryPath(resource)
                    #inform about action
                    oldRelPath = os.path.relpath(oldPath,self.baseDir)
                    newRelPath = os.path.relpath(newPath,self.baseDir)
                    self.logger.Info("Moved resource %s to %s."%(oldRelPath,newRelPath))
                except Exception as e:
                    self.logger.Error("Failed to moved resource: %s"%(str(e)))
                    traceback.print_exc()
                
    def __DeleteResource(self,resource):
        lastPersistentPath = self.__GetLastPersitentPath(resource)
        if(lastPersistentPath):
            try:
                # add deletion to eventsExchangeList
                self.__AddEventToEventsExchangeList(Path(lastPersistentPath),"delete")
                oldPath = lastPersistentPath
                #delete resource
                os.remove(oldPath)
                self.__DeleteResourceOrDirectoryPath(resource)
                #inform about action
                oldRelPath = os.path.relpath(oldPath,self.baseDir)
                self.logger.Info("Deleted resource %s."%(oldRelPath))
            except Exception as e:
                self.logger.Info("Failed to delete resource %s."%(str(e)))
                traceback.print_exc()
                
    ''' directory sync functions '''
    def __AddDirectory(self,parent,directory):
        if(parent): #otherwise there is no need to move the resource
            try:
                newPath = self.GetAbsElementPath(directory)
                if(newPath): #if this is false, then the dir was added to a directory not attached to the workspace
                    # create  in the events exchange list to tell watchdog that the PyEoq created the directory
                    self.__AddEventToEventsExchangeList(Path(newPath), "create")
                    #create new dir file
                    os.mkdir(newPath)
                    self.__SetLastPersistentPath(directory, newPath)
                    #recursively consider all contained resources and subdirs 
                    for res in directory.resources:
                        self.__AddResource(directory, res)
                    for subdir in directory.subdirectories:
                        self.__AddDirectory(directory, subdir)
                    # inform about action
                    newRelPath = os.path.relpath(newPath,self.baseDir)
                    self.logger.Info("Added directory %s."%(newRelPath))
            except Exception as e:
                # remove entry for add directory from events exchange list when it has been created and new path exists
                if (newPath):
                    self.__RemoveEventFromEventsExchangeList(Path(newPath), "create")
                self.logger.Error("Failed to add directory: %s"%(str(e)))
                traceback.print_exc() 
                
    
    def __RenameDirectory(self,directory,newName):
        lastPersistentPath = self.__GetLastPersitentPath(directory)
        if(lastPersistentPath): #otherwise this is a new resource and needs no renaming
            oldPath = lastPersistentPath
            filePath, oldName = os.path.split(oldPath)
            if(newName!=oldName):
                try:
                    newPath = os.path.join(filePath,newName)
                    # make entry into eventsExchangeList to inform watchdog about changes
                    self.__AddEventToEventsExchangeList(Path(newPath),'rename')
                    #create the new dir
                    os.rename(oldPath,newPath) 
                    self.__RefreshResourceOrDirectoryPath(directory)
                    #inform about action
                    oldRelPath = os.path.relpath(oldPath,self.baseDir)
                    newRelPath = os.path.relpath(newPath,self.baseDir)
                    self.logger.Info("Renamed directory %s to %s."%(oldRelPath,newRelPath))
                except Exception as e:
                    # remove entry from events exchange list because change has not been made
                    if self.__IsEventInEventsExchangeList(Path(newPath),'rename'):
                        self.__RemoveEventFromEventsExchangeList(Path(newPath),'rename')

                    self.logger.Error("Failed to rename directory: %s"%(str(e)))
                    traceback.print_exc() 
                    
    def __MoveDirectory(self,directory):
        lastPersistentPath = self.__GetLastPersitentPath(directory)
        if(lastPersistentPath): #otherwise there is no need to move the directory
            oldPath = lastPersistentPath
            newPath = self.GetAbsElementPath(directory)
            if(newPath!=oldPath):
                try:
                    #move the directory 
                    shutil.move(oldPath,newPath) 
                    self.__RefreshResourceOrDirectoryPath(directory)
                    #inform about action
                    oldRelPath = os.path.relpath(oldPath,self.baseDir)
                    newRelPath = os.path.relpath(newPath,self.baseDir)
                    self.logger.Info("Moved directory %s to %s."%(oldRelPath,newRelPath))
                except Exception as e:
                    self.logger.Error("Failed to move directory: %s"%(str(e)))
                    traceback.print_exc()
                
    def __DeleteDirectory(self,directory):
        lastPersistentPath = self.__GetLastPersitentPath(directory)
        if(lastPersistentPath):
            try:
                # add entry to events exchange list to show watchdog that change is made by eoq
                self.__AddEventToEventsExchangeList(Path(self.__GetLastPersitentPath(directory)), "delete")
                oldPath = lastPersistentPath
                #delete resource
                shutil.rmtree(oldPath)
                self.__DeleteResourceOrDirectoryPath(directory)
                #inform about action
                oldRelPath = os.path.relpath(oldPath,self.baseDir)
                self.logger.Info("Deleted directory %s."%(oldRelPath))
            except Exception as e:
                self.logger.Info("Failed to delete directory %s."%(str(e)))
                traceback.print_exc()
                
    ''' Generic path functions '''
                
    def __RefreshResourceOrDirectoryPath(self,element):
        newPath = self.GetAbsElementPath(element)
        if(newPath):
            self.__SetLastPersistentPath(element, newPath)
            if(isinstance(element, ModelResource)):  
                #update the pyecore URI
                newUri = URI(newPath)
                eResource = self.__GetEResourceForModelResource(element)
                eResource.uri = newUri
            elif(isinstance(element, Directory)):
                #update contained elements
                for subres in element.resources:
                    self.__RefreshResourceOrDirectoryPath(subres)
                for subdir in element.subdirectories:
                    self.__RefreshResourceOrDirectoryPath(subdir)
                    
    def __DeleteResourceOrDirectoryPath(self,element):
        self.__DeleteLastPeristentPath(element)
        if(isinstance(element, FileResourceA)):  
            self.__SetResourceClean(element)
            if(element.type == FileResourceTypesE.MODEL):
                self.__UnlinkModelResourceAndEResource(element)
        elif(isinstance(element, Directory)):
            #update contained elements
            for subres in element.resources:
                self.__DeleteResourceOrDirectoryPath(subres)
            for subdir in element.subdirectories:
                self.__DeleteResourceOrDirectoryPath(subdir)
                
    def GetElementPath(self,directory):
        workspaceFound = False #indicates if the element is attached correctly somewhere below the root
        path = directory.name
        if(isinstance(directory,Workspace)):
            return path
        parent = directory.eContainer()
        while(isinstance(parent,Directory)):
            path = os.path.join(parent.name,path)
            if(isinstance(parent,Workspace)):
                workspaceFound = True
                break #exit because there should not be an element further down.
            parent = parent.eContainer()
        if(workspaceFound):
            return path
        else:
            return None
    
    def GetAbsElementPath(self,directory):
        path = self.GetElementPath(directory)
        if(path):
            path = path[2:]
            return os.path.join(self.baseDirAbs,path)
        else:
            return None
     
    def __SetObjectDirty(self,obj):
        if(obj): 
            #find the resource object and mark it dirty
            if(obj in self.dirtyObjects):
                return  # quit here if object is already marked as dirty.
            self.dirtyObjects.append(obj)
            resource = self.__GetResourceForObject(obj)
            self.__SetResourceDirty(resource)
       
    ''' Last persistent path functions '''     
    def __GetLastPersitentPath(self,dirOrResource):
        lastPersistentPath = None
        try:
            lastPersistentPath = self.lastPersistentPaths[dirOrResource]
        except KeyError:
            lastPersistentPath = None
        return lastPersistentPath
    
    def __SetLastPersistentPath(self,dirOrResource,lastPersistentPath):
        self.lastPersistentPaths[dirOrResource] = lastPersistentPath
        
    def __DeleteLastPeristentPath(self,dirOrResource):
        try:
            self.lastPersistentPaths.pop(dirOrResource)
        except:
            pass  # fail silent if element does not has no path so far
    
    ''' Saving functions '''
    def __InitSave(self):
        #stop the current save action 
        self.__StopDelayedSaving()
        if(0<len(self.dirtyResources)):
            if(self.saveTimeout > 0):
                self.__StartDelayedSaving()
            else:
                self.__SaveAllDirtyResources() #skip threading for a 0 timeout
        
    
    def __StartDelayedSaving(self):
        self.saveTimer = Timer(self.saveTimeout,self.__DelayedSaving)
        self.saveTimer.start()
        
    def __StopDelayedSaving(self):
        if(self.saveTimer):
            self.saveTimer.cancel()
            self.saveTimer = None
            
    def __DelayedSaving(self):
        self.__SaveAllDirtyResources()
        self.saveTimer = None
        
    def __SaveAllDirtyResources(self):
        #self.Lock()
        self.mdb.Lock()
        evts = []
        self.__ConnectAllEResources()  # necessary for model resources
        for resource in self.dirtyResources:
            try:
                # add entry to the modified event list to ensure the resource wont get reloaded
                self.AddFileEventToSkipList(FileSystemEvent(self.__GetLastPersitentPath(resource)),2)
                self.__SaveResource(resource)
                evts.append(MsgEvt("WorkspaceMdb","Saved %s."%(resource.name)))
            except Exception as e:
                evts.append(MsgEvt("WorkspaceMdb","Failed to save %s: %s"%(resource.name,str(e))))
                traceback.print_exc()
        self.__DisconnectAllEResources()  # necessary for model resources

        self.dirtyResources.clear()
        self.dirtyObjects.clear()
        
        #notify observers
        self.domain.NotifyObservers(evts, self)
        #self.Release()
        self.mdb.Release()
    
    ''' resource handling functions '''

    def __SetResourceDirty(self,resource):
        if(resource and resource not in self.dirtyResources):
            self.dirtyResources.append(resource)
            self.__InitSave()
            
    def __SetResourceClean(self,resource):
        try:
            self.dirtyResources.remove(resource)
        except:
            pass  # do nothing if the resource was never marked as dirty
 
    def __GetResourceForObject(self,obj):
        container = obj
        while container and not isinstance(container, FileResourceA):
            container = container.eContainer()
        return container
         
    def __LoadMetaModels(self):
        #look for existing meta models and load them on them
        for mp in self.metaDir:
            metadir = mp
            if(not os.path.isabs(mp)):
                metadir = os.path.join(self.baseDir,mp)
            searchString = os.path.join(metadir+'/*.ecore')
            modeldefinitions = glob.glob(searchString)
            for md in modeldefinitions:
                self.__LoadMetaModelResource(md)
        return


    def __LoadModelContents(self,resourceFile):
        resource = self.rset.get_resource(resourceFile)
        return list(resource.contents) #make a copy of the list contents


    def __LoadMetaModelResource(self,resourceFile):
        modelContents = self.__LoadModelContents(resourceFile)
        gcmRoot = modelContents[0]
        self.rset.metamodel_registry[gcmRoot.nsURI] = gcmRoot
        # register all possible subpackages
        for child in gcmRoot.eAllContents():
            if(isinstance(child,EPackage)):
                self.rset.metamodel_registry[child.nsURI] = child
        #remember the file extension, which is the file name (without extension) of the model file
        metaModelFileName = os.path.basename(resourceFile)
        modelExtension = os.path.splitext(metaModelFileName)[0]
        self.knownModelExtensions += [modelExtension]
        #return the root
        return modelContents
                        
            
    def __LoadResourceTree(self):
        (directories,modelfiles) = self.__ScanForFilesAndDirectories()

        #create directories (this includes empty ones)
        for dirpath in directories:
            directory = self.__GetOrCreateDir(dirpath)
        
        #create model files
        for modelfile in modelfiles:
            head,tail = os.path.split(modelfile)
            directory = self.__GetOrCreateDir(head)
            resource = self.__LoadResource(modelfile)
            if(resource): #check if loading succeeded
                resource.name = tail
                directory.resources.add(resource)
    
    
    def __GetOrCreateDir(self,path):
        relPath = os.path.relpath(path, self.baseDirAbs)
        directory = self.modelroot
        if(relPath and relPath != '.'): #only proceed for non empty strings
            segments = relPath.split(os.path.sep)
            for segment in segments: 
                subdirexists = False
                for subdir in directory.subdirectories:
                    if(subdir.name == segment):
                        directory = subdir
                        subdirexists = True
                        break
                if(not subdirexists):
                    newsubdir = Directory(name=segment)
                    self.__SetLastPersistentPath(newsubdir, path)
                    directory.subdirectories.add(newsubdir)
                    directory = newsubdir
        return directory


    def __ScanForFilesAndDirectories(self):
        directories = []
        modelFiles = []
        for root, dirs, files in os.walk(self.baseDirAbs, topdown=True):
            for d in dirs:
#                 relativeRoot = os.path.relpath(root, self.baseDir)
                path = os.path.join(root,d)
                directories.append(path)
            for f in files:
#                 for extension in self.knownModelExtensions:
#                     if(f.endswith('.%s'%(extension))):
#                         relativeRoot = os.path.relpath(root, self.baseDir)
                        path = os.path.join(root,f)
                        modelFiles.append(path)
        return (directories,modelFiles)
    
    def __LoadResource(self,path):
        resource = None
        try:
            extension = os.path.splitext(path)[1].replace('.','') #remove the point in the extension
            self.logger.Info("Loading %s ..."%(path))
            start = default_timer() 
            if(extension in self.knownModelExtensions):
                resource = self.__LoadModelResource(path)
            elif(extension in self.knownXmlExtensions):
                resource = self.__LoadXmlResource(path)
            #check if a handler was found
            if(resource):
                self.__SetLastPersistentPath(resource, path)
                end = default_timer()
                self.logger.Info("ok (%f s)"%(end-start))
            else:
                self.logger.Info("skipped: Unknown format.")
        except Exception as e:
            self.logger.Info("failed: %s"%(str(e)))
        return resource
      
    def __LoadModelResource(self,path):
        resource = ModelResource()
        eResource = self.rset.get_resource(path)
        for content in eResource.contents:
            resource.contents.add(content)
        #register the resource in the lookup table
        self.__LinkModelResourceAndEResource(resource,eResource)
        #re-add the contents to the resource object because the containment removes it
        #is not necessary any more since the content will be re-added when saved or resolving proxies
#         for content in resource.contents:
#             eResource.append(content)
        return resource
    
    def __LoadXmlResource(self,path):
        resource = XmlResource()
        #parse input-xml
        xmlparse = xml.etree.ElementTree
        tree = xmlparse.parse(path)
        root = tree.getroot()
        ldfile = Document(name=resource.name, version="1.0")
        rootelement = Element(name=root.tag, content=root.text.strip())
        resource.document = ldfile
        resource.document.rootelement = rootelement
        for attr in root.attrib:
            newAttr = Attribute(name=attr, value=root.attrib[attr])
            rootelement.attributes.add(newAttr)
        # find first layer
        for element in root:
            if(element.text == None):
                element.text = ""
            #find all subclasses within first layer
            self.__FindChildElements(element, rootelement)
        return resource           
    
    def __FindChildElements(self, element, parent):
        if(element.text == None):
            element.text = ""
        newChild = Element(name = element.tag, content = element.text.strip())
        parent.subelements.add(newChild)
        #create attribute class for each attribute
        for attr in element.attrib:
            newAttr = Attribute(name = attr, value = element.attrib[attr])
            newChild.attributes.add(newAttr)
        #find all child elements
        for child in element:
            self.__FindChildElements(child, newChild)
        return parent
              
    def __SaveResource(self,resource):
        actualPath = self.__GetLastPersitentPath(resource)
        if(actualPath):
            actualRelPath = os.path.relpath(actualPath, self.baseDirAbs)
            self.logger.Info("Saving %s ..."%(actualRelPath))
            start = default_timer()
            if(resource.type == FileResourceTypesE.MODEL): 
                eResource = self.__GetEResourceForModelResource(resource)
                eResource.save()
            elif(resource.type == FileResourceTypesE.XML):
                tree = self.__SaveXmlResource(resource)
                with open(actualPath, "w") as f:
                    #comment: replace is not necessary
                    f.write(tree.replace('<?xml version="1.0" ?>',
                                         '<?xml version="1.0" encoding="utf-8"?>'))  
            elif(resource.type == FileResourceTypesE.TEXT):
                pass
            elif(resource.type == FileResourceTypesE.RAW):
                pass
            #self.__SetLastPersistentPath(resource, actualPath)
            end = default_timer()
            self.logger.Info("ok (%f s)"%(end-start))
        else:
            self.logger.Info("Saving %s skipped,because it is not attached to the workspace."%(resource.name))
        
    def __SaveXmlResource(self, resource):
        ET = xml.etree.ElementTree
        parser = xml.dom.minidom
        #get root element
        rootElement = resource.document.rootelement
        rootTag = ET.Element(rootElement.name)
        rootTag.text = rootElement.content
        #get attributes
        for attrib in rootElement.attributes:
                rootTag.set(attrib.name, attrib.value)
        #find all elements in root
        self.__FindSubElements(ET, rootTag, rootElement)
        #create output string
        xmlstr = parser.parseString(ET.tostring(rootTag)).toprettyxml(indent = "   ")
        return xmlstr   

    def __FindSubElements(self,ET, parentTag, Element):
        for subelement in Element.subelements:
            #add to parent
            subTag = ET.SubElement(parentTag, subelement.name)
            subTag.text = subelement.content
            #get attributes
            for attrib in subelement.attributes:
                subTag.set(attrib.name, attrib.value)
            # find (sub)subelement
            self.__FindSubElements(ET, subTag, subelement)

        
    ''' Special methods for model resources '''
    
    
    def __LinkModelResourceAndEResource(self, modelResource, eResource):
        self.modelResourceLut[eResource] = modelResource
        self.eResourceLut[modelResource] = eResource
        
    def __UnlinkModelResourceAndEResource(self, modelResource):
        try:
            eResource = self.eResourceLut.pop(modelResource)
            self.modelResourceLut.pop(eResource)
        except Exception as e:
            self.logger.Warn("Unlink model resource and eResource failed: %s"%(str(e)))
            
    def __GetModelResourceForEResource(self, eResource):
        try:
            return self.modelResourceLut[eResource]
        except:
            return None
        
    def __GetEResourceForModelResource(self,modelResource):
        try:
            return self.eResourceLut[modelResource]
        except:
            return None
        
    def __ConnectAllEResources(self):
        if(not self.eResourcesConnected):
            #self.eResourceLock.acquire()
            for resourceUri in self.rset.resources:
                resource = self.rset.resources[resourceUri]
                modelResource = self.__GetModelResourceForEResource(resource)
                if(modelResource):
                    for content in resource.contents:
                        resource.remove(content)
                    for content in modelResource.contents:
                        resource.append(content)
                        #modelResource.contents.discard(content)
                    #HACK1: elements must be removed from the Resource in order to make save work
                    modelResource.contents.clear()
            self.eResourcesConnected = True
        
    def __DisconnectAllEResources(self):
        if(self.eResourcesConnected):
            #self.eResourceLock.release()
            for resourceUri in self.rset.resources:
                resource = self.rset.resources[resourceUri]
                modelResource = self.__GetModelResourceForEResource(resource)
                if(modelResource):
                    #HACK2: now read the content
                    for content in resource.contents:
                        modelResource.contents.add(content)
                    #HACK3: content must be re-added to the resource because eResource() wont work otherwise
                    #not needed any more
#                     for content in modelResource.contents:
#                         resource.append(content)
            self.eResourcesConnected = False

    def __CreateCustomEventHandlerForWatchdog(self):
        """Function to create a custom event handler for the watchdog used to track the workspace changes corresponding to .oaam files"""

        def OnCreated(event:FileSystemEvent, pathObject=None):
            # - use path object for the element paths as translator to also support non windows platforms
            #   this if also supports the recursively calling of the OnCreated function with an pathobject
            # - the case when elementPathObject is none is when a copy into the workspace happened and OnCreated event
            #   is thrown for a subfolder whose parent folder has not been created eg. treated in an event.
            #   So the onCreated function gets manually called for this parent folder to create it in advance

            self.logger.Debug('Watchdog: OnCreated')
            if event is None:
                elementPathObject = pathObject
            else:
                elementPathObject = Path(event.src_path)

            # check if the element is known and set element type for later use in the domain command
            elementTypeInEoq = self.__CheckForKnownFile(elementPathObject)
            if not elementTypeInEoq:
                # skip if the file ending is unknown
                return
            # check if element is dir and was already created because previous event of a creation of a child dir caused
            # the creation of this dir. In this case the function is not further needed and can be skipped
            elif elementPathObject.is_dir() and self.__GetHashId(elementPathObject) is not False and \
                    not self.__IsEventInEventsExchangeList(elementPathObject, "create"):
                self.logger.Info(f"Watchdog: Event for creation {str(elementPathObject.absolute())} recognized. Dir already exists. "
                                 f"Skipping event")
                return

            # check if creation was done by the eoq and log if the creation was made by the eoq
            elif self.__CheckAndRemoveFromEventsExchangeList(elementPathObject, "create"):
                return
            else:
                # add the change which has to be made to the change list so there is no interference with the eoq trying
                # to add this folder/file to the file system
                self.__AddEventToEventsExchangeList(elementPathObject, "create")

                # add the skip of the modified event if the event is caused by a file and dont add teh skip event if the
                # onodified function has been called by an other function than the watchdog and when the operating
                # system is mac os
                if elementPathObject.is_file() and event is not None and\
                        "Darwin" not in platform.platform(aliased=True, terse=True):
                    self.AddFileEventToSkipList(event,1)

            # get the parent dir hash ID so it is possible to add the newly created element to it
            parentDirPathObject = elementPathObject.parent
            parentDirHash = self.__GetHashId(parentDirPathObject)

            # if it is not possible to get the parentDirHash, recursively call the OnCreated function
            # (when a onCreated event was thrown for an ).
            # This prevents an  error, when an OnCreation event is thrown for a child dir but the event
            # for its parent dir has not been thrown or worked on, by creating the parent folder in the Eoq in advance.
            if parentDirHash is False:
                self.logger.Info(f"Watchdog: Parent dir '{str(parentDirPathObject.name)}' of '{str(elementPathObject.absolute())}'"
                                 f" does not exist in the Eoq. Recursively calling function OnCreated of the watchdog "
                                 f"to create the parent dir in the Eoq.")
                OnCreated(None,parentDirPathObject)

                # get the parent Dir Hash again
                parentDirHash = self.__GetHashId(parentDirPathObject)

            metaDirAbsPath = self.__GetMetaDirAbsPath()

            # perform changes on the Eoq
            if elementPathObject.is_dir():
                self.__LoadDirFromFileSystem(elementPathObject, parentDirHash)
            # when ecore files are added new files can be added to the eoq, this will be done here
            elif elementPathObject.suffix == ".ecore" and str(parentDirPathObject.absolute()) in metaDirAbsPath:
                # remove entry from list to prevent the exchangeChangeEvents-list to fill up with artifacts or any
                # confusion when the eoq tries to read out an entry out of this list
                self.__LoadMetaModelFromFileSystem(elementPathObject, parentDirPathObject)
            else:
                self.__LoadResourceFromFileSystem(elementPathObject, parentDirPathObject)


        def OnMoved(event:FileSystemEvent):
            self.logger.Debug('Watchdog: OnMoved')
            # get some general information about the object involved in the process
            oldPathToObjectPathElement = Path(event.src_path)
            newPathToObjectPathElement = Path(event.dest_path)
            newNameOfObject = newPathToObjectPathElement.name
            oldNameOfObject = oldPathToObjectPathElement.name
            srcParentDirPathObject = oldPathToObjectPathElement.parent
            destParentDirPathObject = newPathToObjectPathElement.parent

            # this prevents an infinity loop when an revoke of a illegal file suffix renaming made in the explorer
            # was reverted by the watchdog
            if self.__IsEventInEventsExchangeList(newPathToObjectPathElement, 'watchdog_rename'):
                self.__RemoveEventFromEventsExchangeList(newPathToObjectPathElement, 'watchdog_rename')
                self.logger.Info('Watchdog: ok')
                return

            elementTypeInEoq = self.__CheckForKnownFile(oldPathToObjectPathElement)
            if not elementTypeInEoq:
                # check if the file was renamed to a known file suffix and revert the changes
                if newPathToObjectPathElement.suffix[1:] in self.knownXmlExtensions+self.knownModelExtensions:
                    self.logger.Info(f"Watchdog: File {str(oldPathToObjectPathElement.absolute())} was renamed to a file extension, known in"
                                     f" the eoq. Revoking changes...")
                    try:
                        # log changes so the watchdog does not reverte thoose since from his side of view a .oaam file gets illegaly renamed
                        self.__AddEventToEventsExchangeList(oldPathToObjectPathElement, 'watchdog_rename')
                        os.rename(str(newPathToObjectPathElement.absolute()), str(oldPathToObjectPathElement.absolute()))
                    except Exception as e:
                        self.logger.Info("Watchdog: Failed to undo illegal renaming of file %s in the eoq %s."%(str(oldPathToObjectPathElement.absolute()), str(e)))
                        traceback.print_exc()

                    return

                # only add this when the os is windows since there are no onmodified events thrown for this on mac
                if "Darwin" not in platform.platform(aliased=True, terse=True):
                    self.AddFileEventToSkipList(FileSystemEvent(event.dest_path),2)
                return


            ## catch case in which the element is just renamed
            if srcParentDirPathObject.absolute() == destParentDirPathObject.absolute():

                # catch case when a illegal renaming of a file extension happened
                # used != instead of 'is not' because equality and not identity is checked
                if newPathToObjectPathElement.suffix != oldPathToObjectPathElement.suffix:

                    # check whether the changes have been made by the ecoreEditor and delete the event from the event exchange queue
                    # also suffix changes in the eoq are not allowed
                    if self.__CheckAndRemoveFromEventsExchangeList(newPathToObjectPathElement, 'rename'):

                        # add the change which is to be made to the change list so there is no interference with the eoq
                        self.__AddEventToEventsExchangeList(oldPathToObjectPathElement, 'rename')

                        # perform changes
                        self.__RenameResourceOrDir(newPathToObjectPathElement, oldNameOfObject)

                    # add entry to Events exchange list so watchdog does not track the changes itself made
                    self.__AddEventToEventsExchangeList(oldPathToObjectPathElement, 'watchdog_rename')

                    self.logger.Info(f"Watchdog: Illegal renaming of file '{str(oldPathToObjectPathElement.absolute())}' detected. Revoking....")

                    # only add to skip eent list if system is windows
                    if "Darwin" not in platform.platform(aliased=True, terse=True):
                        self.AddFileEventToSkipList(event,1)

                    # undo changes in the explorer
                    os.rename(str(newPathToObjectPathElement),str(oldPathToObjectPathElement))
                    return

                # .ecore is no resource and a renaming should not be done so this is prevented here and changes are reverted
                if oldPathToObjectPathElement.suffix == '.ecore':
                    self.logger.Info(f"Watchdog: Illegal renaming of {str(oldPathToObjectPathElement.absolute())} file detected. Revoking....")
                    # prevent watchdog from reverting changes
                    self.__AddEventToEventsExchangeList(oldPathToObjectPathElement, 'watchdog_rename')
                    os.rename(newPathToObjectPathElement,oldPathToObjectPathElement)
                    return

                # check whether changes have been made by the eoq and return if that is the case
                if self.__CheckAndRemoveFromEventsExchangeList(newPathToObjectPathElement, 'rename'):
                    return

                # make entry in the modifiedqueue to prevent onmodified events to appear for this event, this is done with a synthetic event
                # with the dest_path which is only checked for, if a skip of the onModified event is checked
                # only add to skip list if the os is not mac os
                if not event.is_directory and "Darwin" not in platform.platform(aliased=True, terse=True):
                    self.AddFileEventToSkipList(FileSystemEvent(event.dest_path), 1)

                # add the change which is to be made to the change list so there is no interference with the eoq trying
                # to add this folder to the domain
                self.__AddEventToEventsExchangeList(newPathToObjectPathElement, "rename")
                self.__RenameResourceOrDir(oldPathToObjectPathElement, newNameOfObject)
                return

            # on Windows Platforms the moving of a folder is tracked by the Watchdog with a combination of delete and
            # create events. The only time a OnMoved event is thrown is when a folder or its parent folder is renamed.
            # Renaming its parent folder needs no extra action on the folder itself and this gets caught in here.
            # On OSX there is no combination of delete and create but instead a move event when a folder/file is moved.
            if "Darwin" not in platform.platform(aliased=True, terse=True) and event.is_directory:
                self.logger.Info(f"Watchdog: Moving of '{oldNameOfObject}' from '{str(srcParentDirPathObject)}' to "
                                 f"'{str(destParentDirPathObject)}' detected. No action needed since the moving of one of"
                                 f" its parent folders triggered this event.")
                return

        ## catch case in which the file/folder is moved to another folder, only on mac/linux available s. comment above
            # this catches the case when the Folder/file is moved to an other folder
            elif (newPathToObjectPathElement.exists() and oldPathToObjectPathElement.exists())\
                    or self.__ParentWasMoved(oldPathToObjectPathElement,newPathToObjectPathElement):
                #todo: check if the movement has been made by the eoq is not included. Needs implementation when moving
                #      in the eoq is possible

                # make movement to a create delete sequence to fit mac behavior of watchdog to windows eg make
                # logfiles look the same
                OnDeleted(None, oldPathToObjectPathElement)
                OnCreated(None, newPathToObjectPathElement)

            # catch case when multiply folders are copied
            else:
                elementHash = self.__GetHashId(oldPathToObjectPathElement)
                resource = self.valueCodec.Dec(elementHash)
                self.__SetLastPersistentPath(resource, str(newPathToObjectPathElement.absolute()))
                self.logger.Info(f"Watchdog: Parent folder of {newNameOfObject} renamed. No action needed")

        def OnDeleted(event:FileSystemEvent, pathObject=None):
            self.logger.Debug('Watchdog: OnDeleted')
            # This event behaves different on OSX and Windows when a parent folder is deleted: on windows there is just
            # an event for the deletion of the parent folder. On OSX there is an event for every deleted subfolder in the
            # parent folder. This is caught with an existence check on the parent folder later in this function.

            # make function reusable with a path object
            if event is None:
                elementPathObject = pathObject
            else:
                elementPathObject = Path(event.src_path)

            elementTypeInEoq = self.__CheckForKnownFile(elementPathObject)
            if not elementTypeInEoq:
                return

            # on mac os the modification of an event via an editor is achieved via a create and delete sequence.
            # therefore this sequence will be detected here and treated as an onModified event and the onCreated
            # event will be deleted
            # this sequence should only be searched on files and events which where thrown by the eoq
            if "Darwin" in platform.platform(aliased=True, terse=True) and not elementPathObject.is_dir()\
                    and not event is None:
                evt = self.__observer.event_queue.queue[0]
                if evt[0].src_path == event.src_path and type(evt[0]) is FileCreatedEvent:
                    self.__observer.event_queue.get()
                    OnModified(None, elementPathObject)
                    return

            # check if deletion was made by the eoq
            if self.__CheckAndRemoveFromEventsExchangeList(elementPathObject, 'delete'):
                return

            parentDirPathObject = elementPathObject.parent
            parentDirHash = self.__GetHashId(parentDirPathObject)

            metaDirAbsPath = self.__GetMetaDirAbsPath()

            # perform changes on the local sever domain simply reverting the steps which where made on creation of this
            # elements at server startup
            if elementPathObject.suffix == ".ecore" and str(parentDirPathObject.absolute()) in metaDirAbsPath:
                # delete event from events exchange list since the deletion of the ecore causes no events which could
                # be tracked by the pyeoq
                self.__UnloadMetaModel(elementPathObject)

            elif elementPathObject.name in self.metaDir:
                # get all files which are .ecore Files on top level of the deleted meta folder and unregister them since  only .ecore Files
                # on top level are registerd to track the corresponding models
                elementHash = self.__GetHashId(elementPathObject)
                elementObject = self.valueCodec.Dec(elementHash)

                # get all .ecore files which where contained in the top of the deleted directory and unload them
                for resource in elementObject.resources:
                    if resource.name[-6:] == ".ecore":
                        # convert path so it is absolute to the workspace directory
                        resourceAbsolutePath = self.__GetLastPersitentPath(resource)
                        resourcePathObject = Path(self.baseDir +'\\' +os.path.relpath(resourceAbsolutePath, self.baseDir))
                        self.__UnloadMetaModel(resourcePathObject)

                # unregister Folder and other files in it
                self.__UnloadResourceOrDir(elementPathObject, elementTypeInEoq)

            # in this case the resource is file or dir the procedure of deleting is the same
            else:
                # this case ist just needed for mac since there is no event for deleted subfiles on windows
                if parentDirHash is False:
                    self.logger.Info(
                        f"Watchdog: Parent dir of '{str(elementPathObject)}' already deleted. Performing no changes on the Eoq.")
                else:
                    self.__UnloadResourceOrDir(elementPathObject,elementTypeInEoq)

        def OnModified(event:FileSystemEvent, modifiedObjectPathObject=None):
            self.logger.Debug('Watchdog: OnModified')
            # catch resources which have been changed outside of the ecoreeditor during server runtime
            if event is not None:
                modifiedObjectPathObject = Path(event.src_path)

            # only reload objects corresponding to which fullfill the following conditions:
            # - they are Files, this is tested with event.is_directory and a check if a file ending exists because when the folder is
            # deleted, there is no check if it was a flder with is_file
            # - they are known in the pyeoq via an fitting .ecore file
            # - the corresponding event should not be skipped
            if not modifiedObjectPathObject.is_dir() and not modifiedObjectPathObject.suffix == ""\
                    and self.__CheckForKnownFile(modifiedObjectPathObject,silent=True) and not self.CheckForAndSkipOnModified(event):
                self.logger.Info(f"Watchdog: Modification of File {str(modifiedObjectPathObject.absolute())} detected")
                self.__ReloadResource(modifiedObjectPathObject)


        def OnAnyEvent(event:FileSystemEvent):
            self.logger.Debug('Watchdog: OnAnyEvent')
            if event.event_type == 'modified' and not event.is_directory:
                # onModified events often appear twice for one event so the redundant event gets filtered out here
                time.sleep(0.002)  # wait if eventually a redundant event is occurring
                eventQueueCopy = list(self.__observer.event_queue.queue)
                # search if there is a on modified event which is the same as the current event and delete it from the eventqueue
                for element in eventQueueCopy:
                    if element[0].event_type == 'modified' and element[0].src_path == event.src_path:
                        self.__observer.event_queue.get()
            return

        #### function defininitions are over and the main part of the event handler starts ####

        # add signs so the extensions can be processed in the patternMatchingEventHandler
        fileExtensionsToTrack = ["*"]+["*."+extension for extension in self.knownModelExtensions+self.knownXmlExtensions]

        # set up pattern matching event handler with ignoring temporary files and DS_Store of mac
        # todo: here is a problem when i want all folders to be checked but only specific files
        myEventHandler = PatternMatchingEventHandler(patterns=fileExtensionsToTrack, ignore_patterns=["*.DS_Store", "*.tmp", "*.TMP"],
                                                   case_sensitive=True, ignore_directories=False)

        # pass functions to specific events of the watchdog
        myEventHandler.on_created = OnCreated
        myEventHandler.on_moved = OnMoved
        myEventHandler.on_deleted = OnDeleted
        myEventHandler.on_modified = OnModified
        myEventHandler.on_any_event = OnAnyEvent

        return myEventHandler

    #### useful utility functions are starting here ####
    def __GetHashId(self, srcPath:Path):
        # get the elements of the path and delete the known working directory
        splitPathToObjectTuple = (srcPath.parts)[1:]

        # get the directory of the modelroot
        directory = self.modelroot

        # differentiate between source directory hash, hash of file and hash of dir wanted
        if len(splitPathToObjectTuple) == 0:
            return self.valueCodec.Enc(directory)
        elif "." in splitPathToObjectTuple[-1]:
            file = splitPathToObjectTuple[-1]
            splitPathToObjectTuple = splitPathToObjectTuple[:-1]
        else:
            file = None

        # find the parent object hash in the eoq when splitPathToObjectTuple is at least one element long.
        # Otherwise the root element is the parent object
        if len(splitPathToObjectTuple) >= 1:
            # this try expands the function so it is possible to check if a element exists in the eoq with this function
                # get the directory of the parent folder of the created object by iterate through a part of the
                # array not containing the root element or the destination object
                for element in splitPathToObjectTuple:
                    subdirexists = False
                    for subdir in directory.subdirectories:
                        if subdir.name == element:
                            directory = subdir
                            subdirexists = True
                            break
                    if not subdirexists:
                        return False

        # just return the current workspace as Hash, because there is no subfolder in the path
        if file is None:
            return self.valueCodec.Enc(directory)
        else:
            # get the corresponding resource object of the searched resource and return the object hash
            for resource in directory.resources:
                if resource.name == file:
                    return self.valueCodec.Enc(resource)
            return False

    def __AddEventToEventsExchangeList(self, pathObject:Path, eventId:str):
        '''This function adds an element to the events exchange list so the eoq and the watchdog can communicate about the events they caused'''
        self.exchangeChangeEventsList.append((eventId, pathObject.absolute()))

    def __RemoveEventFromEventsExchangeList(self, pathObject:Path, eventId:str) -> bool:
        if (eventId,pathObject.absolute()) in self.exchangeChangeEventsList:
            self.exchangeChangeEventsList.remove((eventId,pathObject.absolute()))
            return True
        else:
            return False

    def __IsEventInEventsExchangeList(self,pathObject:Path,identifierString:str) -> bool:
        '''Checks if the path element combined with id string is in the events exchange list'''
        for element in self.exchangeChangeEventsList:
            if element[0] == identifierString and element[1].resolve() == pathObject.resolve():
                return True
        return False

    def __CheckAndRemoveFromEventsExchangeList(self, pathObject:Path, identifierString:str, silent=False) -> bool:
        if self.__IsEventInEventsExchangeList(pathObject, identifierString):
            self.__RemoveEventFromEventsExchangeList(pathObject, identifierString)
            if not silent:
                self.logger.Info(f"Watchdog: {identifierString} of {str(pathObject.absolute())} recognized. Skipping Event")
            return True
        else:
            if not silent:
                self.logger.Info(f"Watchdog: {identifierString} of {str(pathObject.absolute())} detected.")
                return False

    def __GetMetaDirAbsPath(self) -> list:
        '''Takes all meta directories and returns all as absolute path strings in an array'''
        metaDirAbsPath = []
        for metaDir in self.metaDir:
            if not os.path.isabs(metaDir):
                metaDirAbsPath.append(os.path.join(self.baseDirAbs, metaDir))
            else:
                metaDirAbsPath.append(metaDir)
        return metaDirAbsPath

    def __LoadMetaModelFromFileSystem(self, elementPathObject:Path, parentDirPathObject=None):
        if parentDirPathObject is None:
            parentDirPathObject = elementPathObject.parent

        # load the ecore model from the files system as normal resource
        self.__LoadResourceFromFileSystem(elementPathObject)

        self.__LoadMetaModelResource(str(elementPathObject.absolute()))

        trackedEnding = str(elementPathObject.name[:-6])

        self.logger.Info(f"Watchdog: Meta Model {elementPathObject} registered.")
        self.logger.Info(f"Watchdog: Scanning and loading {trackedEnding}-Files.")

        # now checking which files have the "new" registered ending and include them into the pyeoq
        # get all the files which are in the workspace
        (directories, modelfiles) = self.__ScanForFilesAndDirectories()

        # counter for later printout and information
        registeredFilesNumber = 0
        notRegisteredFilesNumber = 0

        for element in modelfiles.copy():
            # only work on files which have the now available file ending
            if element[len(element)-len(trackedEnding):] == trackedEnding:

                # configure the filepaths etc for the new file to register
                newFoundFilePathObject = Path(os.path.relpath(element, self.baseDirAbs[:-len(self.baseDir)]))
                newFoundFileParentPathObject = newFoundFilePathObject.parent
                newFoundFileParentDirhash = self.__GetHashId(newFoundFileParentPathObject)

                # add the change which is to be made to the change list so there is no interference with the eoq trying
                # to add this folder/file to the file system
                self.__AddEventToEventsExchangeList(newFoundFilePathObject, "create")

                # uses the function which is also used during the start of the server to ensure compatibility of the changes
                resource = self.__LoadResource(element)

                # check if loading succeeded
                if (resource):
                    resource.name = str(newFoundFilePathObject.name)
                    # get the eoq hash for the loaded resource so it is possible to append the created model to the folder
                    # in the eoq. This is needed to be done with a domain.Do since only those are tracked by the EcoreEditor
                    objectHash = self.valueCodec.Enc(resource)
                    self.domain.Do(Add(newFoundFileParentDirhash, 'resources', objectHash))
                    registeredFilesNumber += 1
                else:
                    self.logger.Info(f"Watchdog: loading of created File '{str(newFoundFilePathObject.name)}' failed. Skipping file.")
                    # remove entry in the events exchange list so there is no confusion and leftover artifact
                    self.__RemoveEventFromEventsExchangeList(elementPathObject, "create")
                    notRegisteredFilesNumber += 1

        self.logger.Info(f"Watchdog: {notRegisteredFilesNumber+registeredFilesNumber} {trackedEnding}-Files found after registering. "
                         f"{registeredFilesNumber} loaded into the Workspace. {notRegisteredFilesNumber} could not be loaded.")

    def __UnloadMetaModel(self, elementPathObject:Path):
        self.logger.Info(f"Watchdog: Unloading {str(elementPathObject.absolute())}")
        start = default_timer()

        # unload meta model as resource
        self.__UnloadResourceOrDir(elementPathObject, 'resources')

        # get the file ending which has to be untracked
        untrackedEnding = str(elementPathObject.name[:-6])

        # reverting changes made during loading the meta model
        self.knownModelExtensions.remove(untrackedEnding)
        modelContents = self.__LoadModelContents(str(elementPathObject.absolute()))
        gcmRoot = modelContents[0]

        # unregister all subpackages belonging to the deleted ecore file
        for child in gcmRoot.eAllContents():
            if (isinstance(child, EPackage)):
                self.rset.metamodel_registry.pop(child.nsURI)
        # unregister meta model file itselve
        self.rset.metamodel_registry.pop(gcmRoot.nsURI)

        end = default_timer()
        self.logger.Info("ok (%f s)"%(end-start))
        self.logger.Info(f"Watchdog: Unloading files corresponding with the deleted meta model")

        # get all the files which are in the workspace
        (directories, modelfiles) = self.__ScanForFilesAndDirectories()

        for element in modelfiles.copy():
            # only work on files which have the now available file ending
            if element[len(element)-len(untrackedEnding):] == untrackedEnding:
                fileToUnloadPathObject = Path(os.path.relpath(element, self.baseDirAbs[:-len(self.baseDir)]))
                self.__UnloadResourceOrDir(fileToUnloadPathObject, 'resources')


        self.logger.Info(f"Watchdog: All files corresponding to {str(elementPathObject.absolute())} unregistered")

    def __LoadResourceFromFileSystem(self, elementPathObject:Path, parentDirPathObject=None):
        # make function usable without parent dir path object
        if parentDirPathObject is None:
            parentDirPathObject = elementPathObject.parent

        self.logger.Info(f"Watchdog: New created file '{str(elementPathObject.name)}' in '{parentDirPathObject.absolute()} detected.'")

        # uses the function which is also used during the start of the server to ensure compatibility of the changes with the eoq and support
        # loading known xml files and known model files
        resource = self.__LoadResource(str(elementPathObject.absolute()))

        # check if loading succeeded
        if (resource):
            resource.name = str(elementPathObject.name)
            # get the eoq hash for the loaded resource so it is possible to append the created model to the folder
            # in the eoq. This is needed to be done with a domain.Do since only those are tracked by the EcoreEditor
            objectHash = self.valueCodec.Enc(resource)
            parentDirHash = self.__GetHashId(parentDirPathObject)
            self.domain.Do(Add(parentDirHash, 'resources', objectHash))
        else:
            self.logger.Info(f"Watchdog: loading of created File '{str(elementPathObject.name)}' failed. Skipping file.")
            # remove entry in the events exchange list so there is no confusion and leftover artifact
            self.__RemoveEventFromEventsExchangeList(elementPathObject, "create")

    def __LoadDirFromFileSystem(self, elementPathObject:Path, parentDirHash):
        # imitate the loading function for directories on server startup to ensure compatibility of the changes
        createdDirectory = Directory(name=str(elementPathObject.name))
        self.__SetLastPersistentPath(createdDirectory, str(elementPathObject.absolute()))

        # the creation of the hash and the operation on the domain is needed so the ecoreeditor is informed
        # about the changes
        directoryHash = self.valueCodec.Enc(createdDirectory)
        self.domain.Do(Add(parentDirHash, 'subdirectories', directoryHash))

    def __UnloadResourceOrDir(self, elementPathObject:Path, elementTypeInEoq=None):
        '''Unload a resource from the eoq without deleting it. Used for the Watchdog'''
        if not elementPathObject.suffix == "":  #this is there to have consistency in the output, with the output when the server starts
            self.logger.Info(f"Watchdog: Unloading {str(elementPathObject.absolute())}....")
        start = default_timer()
        deletedObjectHash = self.__GetHashId(elementPathObject)

        # option to just unload resource locally if the eoq requires no change for example when the parent folder in the
        # eoq is already moved
        # get hashes for later operation on the eoq
        parentDirPathObject = elementPathObject.parent
        parentDirHash = self.__GetHashId(parentDirPathObject)
        # add event to exchange events list so the eoq knows that the change has been made by the Watchdog
        self.__AddEventToEventsExchangeList(elementPathObject, "delete")
        # perform changes on the domain
        self.domain.Do(Cmp().Rem(parentDirHash, elementTypeInEoq, deletedObjectHash))
        try:
            resource = self.valueCodec.Dec(deletedObjectHash)
            self.__DeleteResourceOrDirectoryPath(resource)
            end = default_timer()
            if not elementPathObject.suffix == "":  # this is there to have consistency in the output, with the output when the server starts
                self.logger.Info("ok (%f s)"%(end-start))
        except Exception as e:
            self.logger.Info("Watchdog: Failed to delete resource in the eoq %s."%(str(e)))
            traceback.print_exc()

    def __RenameResourceOrDir(self, elementPathObject:Path, name):
        # get hash id to undo renaming
        objectHash = self.__GetHashId(elementPathObject)

        # perform changes on the eoq
        self.domain.Do(Set(objectHash, 'name', name))

        # refresh path of object in the eoq
        renamedObjectEObject = self.valueCodec.Dec(objectHash)
        self.__RefreshResourceOrDirectoryPath(renamedObjectEObject)

    def __ReloadResource(self, resourcePathObject:Path):
        resourcePath = str(resourcePathObject.absolute())
        if resourcePathObject.suffix[1:] in self.knownModelExtensions:
            resource = self.valueCodec.Dec(self.__GetHashId(resourcePathObject))
            eResource = self.__GetEResourceForModelResource(resource)
            try:
                self.logger.Info("Watchdog: Reloading %s ..." % resourcePath)
                start = default_timer()
                # delete eResource and clean up connection to the resource
                self.__DeleteLastPeristentPath(eResource)
                self.__UnlinkModelResourceAndEResource(resource)
                self.rset.remove_resource(eResource)
                # load new eResource
                newEResource = self.rset.get_resource(resourcePath)
                # swap content
                resource.contents[0] = newEResource.contents[0]
                # register the resource in the lookup table
                self.__LinkModelResourceAndEResource(resource, newEResource)
                self.__SetLastPersistentPath(newEResource, resourcePath)
                # read the contents to the resource object because the containment removes it
                for content in resource.contents:
                    eResource.append(content)
                end = default_timer()
                self.logger.Info("ok (%f s)"%(end-start))
            except Exception as e:
                self.logger.Info("failed: %s"%(str(e)))
        elif resourcePathObject.suffix[1:] in self.knownXmlExtensions:
            resource = self.valueCodec.Dec(self.__GetHashId(resourcePathObject))
            try:
                self.logger.Info("Watchdog: Reloading %s ..." % resourcePath)
                start = default_timer()
                # parse input-xml
                xmlparse = xml.etree.ElementTree
                tree = xmlparse.parse(resourcePath)
                root = tree.getroot()
                ldfile = Document(name=resource.name, version="1.0")
                rootelement = Element(name=root.tag, content=root.text.strip())
                resource.document = ldfile
                resource.document.rootelement = rootelement
                for attr in root.attrib:
                    newAttr = Attribute(name=attr, value=root.attrib[attr])
                    rootelement.attributes.add(newAttr)
                # find first layer
                for element in root:
                    if (element.text == None):
                        element.text = ""
                    # find all subclasses within first layer
                    self.__FindChildElements(element, rootelement)
                end = default_timer()
                self.logger.Info("ok (%f s)"%(end-start))
            except Exception as e:
                self.logger.Info("failed: %s"%(str(e)))

        else:
            self.logger.Info(f"skipped Reloading {resourcePath}: Unknown format.")

    def __CheckForKnownFile(self,elementPathObject:Path, silent = False):
        # check what for element was changed and set a element type so the deletion command can be used for all elements
        if not elementPathObject.suffix:
            elementTypeInEoq = "subdirectories"
        elif elementPathObject.suffix[1:] in self.knownModelExtensions+self.knownXmlExtensions:
            elementTypeInEoq = 'resources'
        else:
            if not silent:
                self.logger.Info(f"The Object '{str(elementPathObject.absolute())}' has no equivalent on the eoq. "
                                 "No action has been performed on the Eoq.")
            return False
        return elementTypeInEoq

    def __ParentWasMoved(self, oldElementPathObject: Path, newElementPathObject: Path) -> bool:
        """ This function decides whether a one of the parent folders of the File/Folder was renamed or was moved
        to another folder """
        workspacePathObject = Path(self.baseDir)
        splitOldPath = list(oldElementPathObject.relative_to(workspacePathObject.parent).parts)
        splitNewPath = list(newElementPathObject.relative_to(workspacePathObject.parent).parts)

        # get the first element from top down in which the two paths differ to check if thoose exist. If both exist the
        # file/folder was just moved. If the old one does not exist, one of hte parent folders was renamed
        for oldPathElement, newPathElement in zip(reversed(splitOldPath), reversed(splitNewPath)):
            if oldPathElement != newPathElement:
                oldDifferentPathObject = oldElementPathObject
                newDifferentPathObject = newElementPathObject
                while oldDifferentPathObject.name != oldPathElement:
                    oldDifferentPathObject = oldDifferentPathObject.parent
                while newDifferentPathObject.name != newPathElement:
                    newDifferentPathObject = newDifferentPathObject.parent
                return newDifferentPathObject.exists() and oldDifferentPathObject.exists()
                pass
            pass
        pass

    def CheckForAndSkipOnModified(self,event:FileSystemEvent) -> bool:
        """Check if the event source path is already marked to skip in the skip list and decrement the skip counter if this is true"""
        for [eventToSkip, timesToSkip] in self.modifiedEventSkipList.copy():
            if eventToSkip.src_path == event.src_path or str(Path(event.src_path).absolute()) == eventToSkip.src_path:
                self.modifiedEventSkipList.remove([eventToSkip, timesToSkip])
                if timesToSkip > 1:
                    self.AddFileEventToSkipList(eventToSkip,timesToSkip-1)
                    return True
                elif timesToSkip == 1:
                    return True
                else:
                    print('Error in the CheckForAndSkipOnModified function of the watchdog. An entry has the'
                          'value 0 for times to skip.')
        return False
    
    def AddFileEventToSkipList(self,fileEvent,n):
        if(self.trackFileChanges):
            self.modifiedEventSkipList.append([fileEvent,n])
            
    def Lock(self):
        self.__ConnectAllEResources()
        #self.mdb.Lock()

    def Release(self):
        self.__DisconnectAllEResources()
        #self.mdb.Release()



'''
 Bjoern Annighoefer 2019
'''

from ..event import EvtProvider,ALL_EVENT_TYPES

class FrameHandler(EvtProvider):
    def __init__(self):
        self.observers = {}
    
    def Handle(self,frames,sessionId=None):
        pass
    
    def Gby(self,sessionId):
        pass
    
    #Override the event provid methods since the sole event provider shall be the cmd runner
    #@Override
    def Observe(self,callback,eventTypes=ALL_EVENT_TYPES,context=None,sessionId=None): #by default register for all events
        pass #needs to be overwritten in frame handler implementation
    
    #@Override    
    def Unobserve(self,callback,context=None):
        pass #needs to be overwritten in frame handler implementation